#include <math.h>
#include "togl.h"

#include "list.h"
#include "routine.h"
#include "track.h"
#include "body.h"		// only for body->number in .i

extern char *trackName[];
extern Tcl_Interp *currentInterp;
extern char error[];

int Figure::s_serial = 1;
int Note::s_serial = 1;

//================================================================================================

Modif::Modif()
{
	type = 0;
	time = -1;
	data.value = data.rightValue = data.leftValue = 0;
	data.rightTime = data.leftTime = -1;
	oldData = data;
}

Modif::Modif(const Modif *modif)
{
	time = modif->time;
	type = modif->type;
	data = modif->data;
	oldData = modif->oldData;
//fprintf(stderr, "duplicate = ");Dump();fprintf(stderr, "\n");
}

Modif::Modif(int flags, int t, double v, double rv, int rt, double lv, int lt)
{
	time = t;
	type = flags;
	if (type & VAL_VALUE) {
		data.value = v;
	}
	if (type & VAL_LEFT) {
		data.leftValue = lv;
		data.leftTime = lt;
	}
	if (type & VAL_RIGHT) {
		data.rightValue = rv;
		data.rightTime = rt;
	}

	oldData.value = oldData.rightValue = oldData.leftValue = 0;
	oldData.rightTime = oldData.leftTime = -1;
}

Modif::~Modif()
{
//fprintf(stderr, "%s  %08x\n", __PRETTY_FUNCTION__, this);
}

//------------------------------------------------------------------------------------------------

void Modif::Save(FILE *out) const
{
int first = TRUE;

	if (type & VAL_VALUE) {
		fprintf(out, "%s%g", first ? "":"+", data.value);
		first = FALSE;
	}
	if ((type & VAL_LEFT) && (type & VAL_RIGHT)) {
		fprintf(out, "%shandles(%g,%d, %g,%d)", first ? "":"+", 
				data.leftValue, data.leftTime, data.rightValue, data.rightTime);
		first = FALSE;
	} else if (type & VAL_LEFT) {
		fprintf(out, "%shandleleft(%g,%d)", first ? "":"+", data.leftValue, data.leftTime);
		first = FALSE;
	} else if (type & VAL_RIGHT) {
		fprintf(out, "%shandleright(%g,%d)", first ? "":"+", data.rightValue, data.rightTime);
		first = FALSE;
	}

	fprintf(out, "@%d", time);
}

//------------------------------------------------------------------------------------------------

#ifdef DEBUG
void Modif::Dump() const
{
int first = TRUE;

	if (type & VAL_VALUE) {
		fprintf(stderr, "%s%g", first ? "":"+", data.value);
		first = FALSE;
	}
	if ((type & VAL_LEFT) && (type & VAL_RIGHT)) {
		fprintf(stderr, "%shandles(%g,%d, %g,%d)", first ? "":"+", 
				data.leftValue, data.leftTime, data.rightValue, data.rightTime);
		first = FALSE;
	} else if (type & VAL_LEFT) {
		fprintf(stderr, "%shandleleft(%g,%d)", first ? "":"+", data.leftValue, data.leftTime);
		first = FALSE;
	} else if (type & VAL_RIGHT) {
		fprintf(stderr, "%shandleright(%g,%d)", first ? "":"+", data.rightValue, data.rightTime);
		first = FALSE;
	}

	fprintf(stderr, "@%d [%s%g -- %g,%d, %g,%d]", time, type & NEW_KEY ? "new -- " : "",
			oldData.value, oldData.rightValue, oldData.rightTime, 
			oldData.leftValue, oldData.leftTime);
}
#endif

//================================================================================================

Key::Key()
{
	for(int i=0; i<NB_DOF; i++)
		part[i] = 0;

	dancer = 0;
}

Key::~Key()
{
//fprintf(stderr, "%s  %08x\n", __PRETTY_FUNCTION__, this);

	for(int i=0; i<NB_DOF; i++) {
		delete part[i];
	}
}

//------------------------------------------------------------------------------------------------

int Key::Compile(class Interpolator *dancers[])
{
	for (int i=0; i<NB_DOF; i++) {
		if (part[i]) {
//fprintf(stderr, "\ntrack = %-20s  ", trackName[i]);
			for (ListIterator<Chunk> c(part[i]); c; ++c) {
				if (dancers[dancer]->track[i].AddInPlace(c()) == FALSE) {
					sprintf(error, "track '%s' in key(%d)@[%d:%d]",
							trackName[i], dancer, start, end);
					Tcl_AppendResult(currentInterp, error, 0);
					return FALSE;
				}
			}
		}

		delete part[i];
		part[i] = 0;
	}

	return TRUE;
}

//------------------------------------------------------------------------------------------------

// Helper function for syntaxic analysis
int Key::CheckIntegrity()
{
int timeMin = 1000000000, timeMax = -1;

	for (int i=0; i<NB_DOF; i++) {
		if (part[i]) {
			for (ListIterator<Chunk> c(part[i]); c; ++c) {
				if (c()->startTime<timeMin) timeMin = c()->startTime;
				if (c()->endTime>timeMax) timeMax = c()->endTime;
			}
		}
	}

	return (timeMin<start || timeMax>end ? FALSE : TRUE);
}

//------------------------------------------------------------------------------------------------

#ifdef DEBUG
void Key::Dump() const
{
int i;

	fprintf(stderr, "    key(%d) @ [%d:%d]\n", dancer, start, end);

	for (i=0; i<NB_DOF; i++) {
		if (part[i]) {
			fprintf(stderr, "        %s = ", trackName[i]);
			for (ListIterator<Chunk> c(part[i]); c; ++c) {
				c()->Dump();
			}

			part[i]->last()->DumpEnd();
			fprintf(stderr, "\n");
		}
	}
	fprintf(stderr, "\n");
}
#endif

//================================================================================================

Figure::Figure() : number(s_serial++)
{
int i,j;

	for(i=0; i<MAX_DANCERS; i++) {
		for(j=0; j<NB_DOF; j++) {
			modifications[i][j] = 0;
			flags[i][j] = 0;
		}

		map[i] = -1;
	}

	start = end = nbMap = 0;
	name = filename = color = 0;
	includeNotes = FALSE;
}

Figure::~Figure()
{
//fprintf(stderr, "%s  %08x\n", __PRETTY_FUNCTION__, this);

	for(int i=0; i<MAX_DANCERS; i++)
		for(int j=0; j<NB_DOF; j++)
			delete modifications[i][j];

	delete[] filename;
	delete[] name;
	delete[] color;
}

//------------------------------------------------------------------------------------------------

int Figure::RegisterModif(int dancer, int track, class Modif *modif)
{
ListIterator<Modif> m;
int found = FALSE;

	// if track rejected or not used, skip it
	if ((flags[dancer][track] & REJECT) || !(flags[dancer][track] & USED)) {
		return 0;
	}

	// test if modification is in this figure
	if (modif->time < start || end < modif->time) {
		return 0;
	}

//fprintf(stderr, "-----> modif = ");modif->Dump();fprintf(stderr, "\n");
	// register it
	if (modifications[dancer][track]) {
		for(m.reset(modifications[dancer][track]); m; ++m) {
			if (m()->time == modif->time && !(modif->type & TEST_ONLY)) {
//fprintf(stderr, "-----> updating = ");m()->Dump();fprintf(stderr, "\n");
				// set the possible oldData struct to new modif before replacing it
				if (modif->type & VAL_VALUE) {
					if (!(m()->type & OLD_VALUE)) {
//fprintf(stderr, "-----> setting oldDATA -- VAL_VALUE\n");
						m()->oldData.value = modif->oldData.value;
						m()->type |= OLD_VALUE;
					}
					m()->data.value = modif->data.value;
					m()->type |= VAL_VALUE;
				}
				// left    (if not at the end of figure)
				if ((modif->type & VAL_LEFT) && (modif->time != end)) {
					if (!(m()->type & OLD_LEFT)) {
//fprintf(stderr, "-----> setting oldDATA -- VAL_LEFT\n");
						m()->oldData.leftValue = modif->oldData.leftValue;
						m()->oldData.leftTime = modif->oldData.leftTime;
						m()->type |= OLD_LEFT;
					}
					m()->data.leftValue = modif->data.leftValue;
					m()->data.leftTime = modif->data.leftTime;
					m()->type |= VAL_LEFT;
				}
				// right    (if not at the start of figure)
				if ((modif->type & VAL_RIGHT)  && (modif->time != start)) {
					if (!(m()->type & OLD_RIGHT)) {
//fprintf(stderr, "-----> setting oldDATA -- VAL_RIGHT\n");
						m()->oldData.rightValue = modif->oldData.rightValue;
						m()->oldData.rightTime = modif->oldData.rightTime;
						m()->type |= OLD_RIGHT;
					}
					m()->data.rightValue = modif->data.rightValue;
					m()->data.rightTime = modif->data.rightTime;
					m()->type |= VAL_RIGHT;
				}
				// we have found it
				found = TRUE;
				break;
			}
		}
	}

	if (!found && !(modif->type & TEST_ONLY)) {
		if (!modifications[dancer][track]) {
			modifications[dancer][track] = new List<Modif>;
		}

		class Modif *m = new Modif(modif);

		if (m->type & VAL_VALUE) {
			m->oldData.value = modif->oldData.value;
			m->type |= OLD_VALUE;
		}
		if (m->type & VAL_LEFT) {
			m->oldData.leftValue = modif->oldData.leftValue;
			m->oldData.leftTime = modif->oldData.leftTime;
			m->type |= OLD_LEFT;
		}
		if (m->type & VAL_RIGHT) {
			m->oldData.rightValue = modif->oldData.rightValue;
			m->oldData.rightTime = modif->oldData.rightTime;
			m->type |= OLD_RIGHT;
		}
//fprintf(stderr, "-----> creating new modif at %d & setting oldDATA (append)\n", modif->time);
		modifications[dancer][track]->append(m);
	}

//fprintf(stderr, "-----> returning %s\n", found ? "FALSE" : "TRUE");
	if (modif->time == start || modif->time == end) {
		return (found ? 2 : -2);
	} else {
		return (found ? 1 : -1);
	}
}

//------------------------------------------------------------------------------------------------

int Figure::DeleteModif(class Routine &routine, int dancer, int track, int startTime, int endTime)
{
ListIterator<Modif> m;
int found = FALSE;

//fprintf(stderr, "--------------------\nDELETE_MODIF on [%d / %s]  --  range %d:%d  --  figure %d:%d\n", dancer, trackName[track], startTime, endTime, start, end);

	// if track rejected or not used, skip it
	if ((flags[dancer][track] & REJECT) || !(flags[dancer][track] & USED)) {
		return FALSE;
	}

	// test if modification is in this figure
	if (endTime < start || end < startTime) {
		return FALSE;
	}

	// find the modif
	if (modifications[dancer][track]) {
		for(m.reset(modifications[dancer][track]); m; ++m) {
			if (startTime <= m()->time && m()->time <= endTime) {
//fprintf(stderr, "    @@@ found = ");m()->Dump();fprintf(stderr, "\n");
				if (m()->type & NEW_KEY) {
					//--------------------------
					// deletion of a created key
//fprintf(stderr, "    @@@ deletion of NEW_KEY\n");
					routine.dancers[dancer]->track[track].Delete(m()->time, m()->time,
																 routine.nbBeats);
					modifications[dancer][track]->destroy(m());
					if (modifications[dancer][track]->head() == 0) {
						delete modifications[dancer][track];
						modifications[dancer][track] = 0;
					}
					break;					

				} else {
					//----------------------------------------
					// deletion of a value & restore old value
//fprintf(stderr, "    @@@ modify value\n");
					int beforeTime, afterTime;
					class Modif *toDelete1=0, *toDelete2=0; 
					routine.ModifyValue(m()->time, dancer, track, 0, 0, m()->oldData.value,
										&beforeTime, &afterTime);
					for(ListIterator<Modif> mm(modifications[dancer][track]); mm; ++mm) {
						if (mm()->time == beforeTime) {
//fprintf(stderr, "    @@@ reseting before at %d\n", beforeTime);
							mm()->type &= ~VAL_LEFT;
							if ((mm()->type & VAL_MASK) == 0) toDelete1 = mm();
						} else if (mm()->time == afterTime) {
//fprintf(stderr, "    @@@ reseting after at %d\n", afterTime);
							mm()->type &= ~VAL_RIGHT;
							if ((mm()->type & VAL_MASK) == 0) toDelete2 = mm();
						}
					}
					modifications[dancer][track]->destroy(m());
					if (toDelete1) modifications[dancer][track]->destroy(toDelete1);
					if (toDelete2) modifications[dancer][track]->destroy(toDelete2);
					if (modifications[dancer][track]->head() == 0) {
						delete modifications[dancer][track];
						modifications[dancer][track] = 0;
					}
					break;
				}
			}
		}
	}


	return TRUE;
}

//------------------------------------------------------------------------------------------------

// Helper function for syntaxic analysis
int Figure::CheckIntegrity()
{
int timeMin = 1000000000, timeMax = -1;

	for (int i=0; i<MAX_DANCERS; i++) {
		for (int j=0; j<NB_DOF; j++) {
			if (modifications[i][j]) {
				for (ListIterator<Modif> m(modifications[i][j]); m; ++m) {
					if (m()->time<timeMin) timeMin = m()->time;
					if (m()->time>timeMax) timeMax = m()->time;
				}
			}
		}
	}

	return (timeMin<start || timeMax>end ? FALSE : TRUE);
}

//------------------------------------------------------------------------------------------------

int Figure::Compile(class Routine &dest)
{
class Routine routine(-1);
int i;

	//-----------------------------
	// load the file and compile it
	if (routine.Compile(filename, COMPILE) == FALSE) {
		Tcl_AppendResult(currentInterp, "\nwhile compiling '", filename, "'", 0);
		return FALSE;
	}

	// remove the synthetic chunk in figure
	for(i=0; i<routine.nbDancers; i++) {
		routine.dancers[i]->RemoveSynthetic();
	}

	return Insert(dest, routine);
}


int Figure::Insert(class Routine &dest, class Routine &source)
{
int i,j;

	//-------------------
	// copy the name info
	if (source.name) {
		name = new char[strlen(source.name)+1];
		strcpy(name, source.name);
	}

	//---------------------------
	// some checks against errors
	if (nbMap != source.nbDancers) {
		sprintf(error,
				"routine(%s)@[%d:%d] : dancers in map (%d) and in routine (%d) incompatible",
				filename, start, end, nbMap, source.nbDancers);
		Tcl_AppendResult(currentInterp, error, 0);
		return FALSE;
	}
	if (end-start != source.nbBeats) {
		sprintf(error, 
				"routine(%s)@[%d:%d] : time duration (%d) not compatible with routine (%d)", 
				filename, start, end, end-start, filename, source.nbBeats);
		Tcl_AppendResult(currentInterp, error, 0);
		return FALSE;
	}

	//----------------------
	// reset dancer in space
	for(i=0; i<source.nbDancers; i++) {
		// translate it to (0,0)
		source.dancers[i]->track[cogX].JointValue();
		source.dancers[i]->track[cogZ].JointValue();
		// orient it towards 0 deg
//fprintf(stderr, "\n=============== Figure::Insert  %d   source to zero", i);
		source.OrientDancer(i, -1, source.nbBeats);
	}

	//---------------------------
	// translate it to start time
	for(i=0; i<source.nbDancers; i++) {
		source.dancers[i]->TranslateTime(start);
	}

	//-------------------------------------------------------
	// do the mapping and insert the data in the interpolator
	for (i=0; i<nbMap; i++) {
		if (map[i] == -1) {
			// if dancer not used in the mapping
			continue;
		}

		// transfert info form the compiled routine into the current one
		for (j=0; j<NB_DOF; j++) {
			if (!(flags[i][j] & REJECT) && source.dancers[i]->track[j].IsUsed()) {
				flags[i][j] |= USED;
				if (dest.dancers[map[i]]->track[j].AddInPlace(&source.dancers[i]->track[j])
						== FALSE) {
//fprintf(stderr, "\n-- ERROR ------ flags = %d",map[i],trackName[j],flags[i][j]);
					sprintf(error,
							"routine(%s)@[%d:%d] :  in track %s for dancer %d",
							filename, start, end, trackName[j], i);
					Tcl_AppendResult(currentInterp, error, 0);
					return FALSE;
				}

//fprintf(stderr, "\ndancer %d -- %-20s",map[i],trackName[j]);
				dest.dancers[map[i]]->track[j].Smooth(start, end, FALSE);
			}
		}
	}

	//----------------------------------------------------
	// joint the cogX, cogZ and cogRY to orient the dancer
	for(i=0; i<nbMap; i++) {
		if (map[i] == -1)
			continue;

//fprintf(stderr, "\n=============== Figure::Insert(1)  %d  --  %d:%d : %d\n", map[i],start,end,dest.nbBeats);
		dest.OrientDancer(map[i], start, end);
		if (end < dest.nbBeats) {
//fprintf(stderr, "\n=============== Figure::Insert(2)  %d  --  %d:%d\n", map[i],end,dest.nbBeats);
			dest.OrientDancer(map[i], end, dest.nbBeats);
		}
	}

	//------------------------------------
	// apply the changes defined in 'with'

	for(i=0; i<nbMap; i++) {
		for(j=0; j<NB_DOF; j++) {
			if (!(flags[i][j] & REJECT) && modifications[i][j] 
					&& map[i]!=-1 && source.dancers[i]->track[j].IsUsed()) {
				for(ListIterator<Modif> m(modifications[i][j]); m; ++m) {
					if (dest.dancers[map[i]]->track[j].ApplyModif(m(), modifications[i][j])
							== FALSE) {
						sprintf(error, "routine(%s)@[%d:%d] : can't apply modification for "
								"track '%s' for dancer %d at time %d",
								filename, start, end, trackName[j], i, m()->time);
						Tcl_AppendResult(currentInterp, error, 0);
						return FALSE;
					}
				}
			}
		}
	}

	//------------------------
	// include notes in needed
	if (!includeNotes) {
		return TRUE;
	}

	source.SynthetizeNotes(start);
	
	for (i=0; i<nbMap; i++) {
		if (map[i] == -1) {
			// if dancer not used in the mapping
			continue;
		}

		if (source.notes[i][Routine::REAL]) {
			if (dest.notes[map[i]][Routine::INCLUDED] == 0) {
				dest.notes[map[i]][Routine::INCLUDED] = new List<Note>;
			}

			dest.notes[map[i]][Routine::INCLUDED]->append(source.notes[i][Routine::REAL]);
			source.notes[i][Routine::REAL] = 0;
		}
	}


	return TRUE;
}

//------------------------------------------------------------------------------------------------

int trackToJoint[] = {
	bPelvis, bPelvis, bPelvis,
	bTorso, bTorso, bTorso,
	bHead, bHead, bHead,

	lShoulder, lShoulder, lShoulder, lShoulder,
	lElbow, lElbow,
	lWrist, lWrist, lWrist,
	lHip, lHip, lHip,
	lKnee,
	lAnkle, lAnkle, lAnkle,
	lFoot,
	
	rShoulder, rShoulder, rShoulder, rShoulder,
	rElbow, rElbow,
	rWrist, rWrist, rWrist,
	rHip, rHip, rHip,
	rKnee,
	rAnkle, rAnkle, rAnkle,
	rFoot,

	cXZ, cY, cXZ,
	cRX, cRY, cRZ };


void Figure::Save(FILE *out) const
{
extern void Stringify(FILE *out, char *str);
int i,j;
int header;


	fprintf(out, "routine(%s) @ [%d:%d] = {\n", filename, start, end);
	
	fprintf(out, "    map(");
	for(i=0; i<nbMap; i++)
		fprintf(out, "%d%s", map[i], i==nbMap-1 ? ")\n" : ",");

	header = FALSE;
	for(i=0; i<MAX_DANCERS; i++) {
		extern char *jointName[];
		int alreadyDone[NB_JOINTS];
		for(j=0; j<NB_JOINTS; j++)
			alreadyDone[j] = FALSE;
		
		for(j=0; j<NB_DOF; j++) {
			if (flags[i][j] & REJECT) {
				if (!header) { header = TRUE; fprintf(out, "    reject { "); }
				if (!alreadyDone[trackToJoint[j]]) {
					fprintf(out, "%s(%d) ", jointName[trackToJoint[j]],i);
					alreadyDone[trackToJoint[j]] = TRUE;
				}
			}
		}
	}
	if (header) fprintf(out, "}\n");

	header = FALSE;
	for(i=0; i<MAX_DANCERS; i++) {
		for(j=0; j<NB_DOF; j++) {
			if (modifications[i][j]) {
				if (!header) { header = TRUE; fprintf(out, "    with {"); }
				fprintf(out, "\n        %s(%d) = ", trackName[j], i);
				ListIterator<Modif> m(modifications[i][j]);
				m()->Save(out);
				for (++m; m; ++m) {
					fprintf(out, ", ");
					m()->Save(out);
				}
			}
		}
	}
	if (header) fprintf(out, "\n    }\n");

	if (color) {
		fprintf(out, "    color = ");
		Stringify(out, color);
		fprintf(out, "\n");
	}

	if (includeNotes) {
		fprintf(out, "    includenotes\n");
	}

	fprintf(out, "}\n");
}

//------------------------------------------------------------------------------------------------

int Figure::ApplyToDancer(int dancer) const
{
	for(int i=0; i<nbMap; i++) {
		if (map[i] == dancer)
			return TRUE;
	}
	return FALSE;
}

//------------------------------------------------------------------------------------------------

#ifdef DEBUG
void Figure::Dump() const
{
int i,j;

	fprintf(stderr, "    routine(%s) @ [%d:%d]\n", filename, start, end);

	if (color) fprintf(stderr, "        color = %s\n", color);
	fprintf(stderr, "        map(");
	
	for(i=0; i<nbMap; i++)
		fprintf(stderr, "%d%s", map[i], i==nbMap-1 ? ")\n" : ", ");

	fprintf(stderr, "        reject { ");
	for(i=0; i<MAX_DANCERS; i++)
		for(j=0; j<NB_DOF; j++)
			if (flags[i][j] & REJECT)
				fprintf(stderr, "%s(%d) ", trackName[j],i);
	fprintf(stderr, "}\n");

	fprintf(stderr, "        use { ");
	for(i=0; i<MAX_DANCERS; i++)
		for(j=0; j<NB_DOF; j++)
			if (flags[i][j] & USED)
				fprintf(stderr, "%s(%d) ", trackName[j],i);
	fprintf(stderr, "}\n");

	fprintf(stderr, "        with { ");
	for(i=0; i<MAX_DANCERS; i++) {
		for(j=0; j<NB_DOF; j++) {
			if (modifications[i][j]) {
				fprintf(stderr, "\n            %s(%d) = ", trackName[j], i);
				ListIterator<Modif> m(modifications[i][j]);
				m()->Dump();
				for (++m; m; ++m) {
					fprintf(stderr, ", ");
					m()->Dump();
				}
			}
		}
	}
	fprintf(stderr, "}\n");

	fprintf(stderr, "        includenotes = %s\n", includeNotes ? "yes" : "no");
}
#endif
