#include "togl.h"
#include <math.h>
#ifndef WIN32
#include <stdlib.h>		// for abs()
#endif

#include "list.h"
#include "routine.h"
#include "track.h"
#include "movie.h"
#include "body.h"		// only for trackEnum

extern char *trackName[];
extern char *jointName[];
extern Tcl_Interp *currentInterp;
char error[200];

//================================================================================================

Routine::Routine(int counter)
{
	number = counter;

	for(int i=0; i<MAX_DANCERS; i++) {
		dancers[i] = 0;
		bodies[i] = 0;
		dancerNames[i] = 0;
		dancerSex[i] = MAN;
		dancerTypes[i] = 0;
		notes[i][REAL] = notes[i][INCLUDED] = 0;
		for(int j=0; j<6; j++)
			dancerInitPos[i][j] = 0.0;
	}
	notes[COMMENTS][REAL] = notes[COMMENTS][INCLUDED] = 0;
	nbDancers = nbBeats = 0;

	description = name = copyright = text = 0;
	keys = 0;
	figures = 0;
	movies = new List<Movie>;
	time = 0;
	rhythm = 4000;
	bpm = 30;
}

Routine::~Routine()
{
//fprintf(stderr, "%s  %08x\n", __PRETTY_FUNCTION__, this);

	delete[] name;
	delete[] description;
	delete[] copyright;
	delete[] text;
	for(int i=0; i<nbDancers; i++) {
		delete[] dancerNames[i];
		delete[] dancerTypes[i];
		delete dancers[i];
		if (notes[i][REAL]) notes[i][REAL]->free();
		if (notes[i][INCLUDED]) notes[i][INCLUDED]->free();
		delete notes[i][REAL];
		delete notes[i][INCLUDED];
	}
	if (notes[COMMENTS][REAL]) notes[COMMENTS][REAL]->free();
	if (notes[COMMENTS][INCLUDED]) notes[COMMENTS][INCLUDED]->free();

	if (keys) keys->free();
	delete keys;
	if (figures) figures->free();
	delete figures;
	if (movies) movies->free();
	delete movies;
}

//------------------------------------------------------------------------------------------------

#ifdef DEBUG
void Routine::Dump() const
{
int i,j;

	fprintf(stderr, "\n======== Routine =========\n");
	fprintf(stderr, "nb beats   = %d\n", nbBeats);
	fprintf(stderr, "nb dancers = %d\n"
					"names      = ", nbDancers);

	fprintf(stderr, "%s (%s)", dancerNames[0], dancerSex[0] == MAN ? "man" : "woman");
	for(i=1; i<nbDancers; i++)
		fprintf(stderr, ", %s (%s)", dancerNames[i], dancerSex[i] == MAN ? "man" : "woman");

	fprintf(stderr, "\n");
	fprintf(stderr, "descript   = %s\n", description);
	fprintf(stderr, "name       = %s\n", name);

	fprintf(stderr, "\n<init>  = \n");
	for(i=0; i<nbDancers; i++) {
		fprintf(stderr, "    dancer %d = \n",i);
		for(j=cogX; j<=cogRZ; j++) {
			if (dancers[i] && dancers[i]->track[j].IsUsed()) {
				fprintf(stderr, "        %s = %g\n", trackName[j], dancerInitPos[i][j-cogX]);
			}
		}
		fprintf(stderr, "\n");
	}

	fprintf(stderr, "<keys>  = \n");
	for(ListIterator<Key> k(keys); k; ++k) {
		fprintf(stderr, "--------------------------\n");
		k()->Dump();
	}
	fprintf(stderr, "<figures>  = \n");
	for(ListIterator<Figure> f(figures); f; ++f) {
		f()->Dump();
	}

	fprintf(stderr, "\n<notes>  = \n");
	for(i=0; i<MAX_DANCERS+1; i++) {
		ListIterator<Note> n(notes[i][REAL]);
		if (n) {
			fprintf(stderr, "    dancer %d = ",i);
			n()->Dump();
			for(++n; n; ++n) {
				fprintf(stderr, ", ");
				n()->Dump();
			}
		}
		fprintf(stderr, "\n");
	}

	fprintf(stderr, "<movies>  = \n");
	for(ListIterator<Movie> m(movies); m; ++m) {
		m()->Dump();
	}

	fprintf(stderr, "\n==========================\n\n");
	for(i=0; i<nbDancers; i++) {
		fprintf(stderr, "------- DANCER %d ------------------------------------------------\n",i);
		if (dancers[i]) dancers[i]->Dump(TRUE);
	}

	fprintf(stderr, "==========================\n\n");

	List<TimeRange> *times;
	for(i=0; i<nbDancers; i++) {
		fprintf(stderr, "------- DANCER %d -------\n",i);
		for(j=0; j<NB_DOF; j++) {
			if (dancers[i]->track[j].IsUsed()) {
				fprintf(stderr, "    %-20s = ", trackName[j]);
				times = ListKeysTime(i, j);
				for(ListIterator<TimeRange> t(times); t; ++t) {
					fprintf(stderr, "    %d-%d  ", t()->start, t()->end);
				}
				fprintf(stderr, "\n");
				times->free();
				delete times;
			}
		}
	}


	fprintf(stderr, "==========================\n\n");


}
#endif

//================================================================================================

int Routine::Compile(char *filename, int mode)
{
extern FILE *yyin;
extern int yyparse(void);
extern int yydebug;
extern int yylineno;
extern class Routine *routine;
extern int parserMode;


	// 1. read the file
	if ((yyin = fopen(filename, "r")) == 0) {
		char full[256];
		sprintf(full, "%s/%s",
					  Tcl_GetVar2(currentInterp, "status", "dir:routine", TCL_GLOBAL_ONLY),
					  filename);
#ifdef WIN32
		for(char *tmp = full; *tmp != '\0'; tmp++) {
			if (*tmp == '/') {
				*tmp = '\\';
			}
		}
#endif
		if ((yyin = fopen(full, "r")) == 0) {
	        Tcl_AppendResult(currentInterp, "can't open file to load\n", TCL_STATIC);
			return TCL_ERROR;
		}
	}

	// 2. clean the old content of the routine
	delete[] description;
	description = 0;
	delete[] name;
	name = 0;
	for(int i=0; i<nbDancers; i++) {
		delete[] dancerNames[i];
		dancerNames[i] = 0;
		dancerSex[i] = MAN;
		delete dancers[i];
		dancers[i] = 0;
	}
	nbDancers = nbBeats = 0;

	if (keys) keys->free();
	delete keys;
	keys = 0;
	if (figures) figures->free();
	delete figures;
	figures = 0;

	description = 0;
	name = 0;
	if (mode == COMPILE) {
		keys = new List<Key>;
		figures = new List<Figure>;
	}

	// 3. some init
	routine = this;
	parserMode = mode;
	yylineno = 1;

	// 4. parse the file and generate the entries (if in compile mode)
	if (yyparse() != 0) {
		fclose(yyin);
		return FALSE;
	}
	fclose(yyin);

	// 5. compile
	if (mode != COMPILE) {
		return TRUE;
	} else {
		return Generate();
	}
}

//------------------------------------------------------------------------------------------------

int Routine::Generate()
{
ListIterator<Figure> f;
ListIterator<Key> k;
int max = -1, i;

	// 1. check some errors
	if (nbBeats == 0) {
		Tcl_AppendResult(currentInterp,
						 "beats=0 for the routine", 0);
		return FALSE;
	}

	for(f.reset(figures); f; ++f) {
		if (f()->end > max)
			max = f()->end;
	}
	for(k.reset(keys); k; ++k) {
		if (k()->end > max)
			max = k()->end;
	}
	if (max > nbBeats) {
		Tcl_AppendResult(currentInterp,
						 "beats and time range of entries incompatible", 0);
		return FALSE;
	}
	
	// 2. create new interpolator
	for(i=0; i<nbDancers; i++) {
		dancers[i] = new Interpolator;
	}

	// 3. fill them with the real data
	for(k.reset(keys); k; ++k) {
		if (k()->Compile(dancers) == FALSE) {
			return FALSE;
		}
	}
	keys->free();
	delete keys;
	keys = 0;

	for(f.reset(figures); f; ++f) {
		if (f()->Compile(*this) == FALSE) {
			return FALSE;
		}
	}

	// 4. use init position
	for(i=0; i<nbDancers; i++) {
		if (dancerInitPos[i][4] != 0.0) {
//fprintf(stderr, "=============== Routine::Generate  %d  --  %d:%d\n", i, 0, nbBeats);
			OrientDancer(i, 0, nbBeats);
		}
	}

	// 5. remove the gaps
	for(i=0; i<nbDancers; i++) {
		dancers[i]->Smooth(0, nbBeats, TRUE);
	}

	return TRUE;
}

//------------------------------------------------------------------------------------------------

void Routine::NewNbBeats()
{
	for(int i=0; i<nbDancers; i++) {
		dancers[i]->AdjustNbBeats(nbBeats);
	}
}

//------------------------------------------------------------------------------------------------

void Routine::OrientDancer(int dancer, int start, int end)
{
double angle = 0, cosine, sine, tmp;
Track &tX = dancers[dancer]->track[cogX];
Track &tZ = dancers[dancer]->track[cogZ];


//fprintf(stderr, "---------------- ORIENT %d --- [%d:%d] --------------\n",dancer,start,end);

	if (dancers[dancer]->track[cogRY].IsUsed()) {
		// 1. parameters for the orientation
		// 1.1. angle is joint for figure, just translated for init position
		if (start < 0) {
//fprintf(stderr, "\n--------------- start<0  dancer %d , %g <-> %g",dancer,dancers[dancer]->track[cogRY].data->first()->startValue,dancerInitPos[dancer][4]);
			angle = -dancers[dancer]->track[cogRY].TranslateValue(-dancers[dancer]->track[cogRY].data->first()->startValue);
//			start = 0;
		} else if (start == 0) {
//fprintf(stderr, "\n--------------- start==0  dancer %d , %g <-> %g",dancer,dancers[dancer]->track[cogRY].data->first()->startValue,dancerInitPos[dancer][4]);
			if (dancers[dancer]->track[cogRY].data->first()->startValue
					!= dancerInitPos[dancer][4]) {
//fprintf(stderr, "\n--------------- angle(Trans)   dancer %d , %g",dancer,dancerInitPos[dancer][4]);
				angle = -dancers[dancer]->track[cogRY].TranslateValue(dancerInitPos[dancer][4]);
			} else {
//fprintf(stderr, "\n--------------- no angle(Trans)");
			}
		} else {
//fprintf(stderr, "\n--------------- angle(Joint)   dancer %d , joint %d-%d , beats %d",dancer,start,end,nbBeats);
			angle = -dancers[dancer]->track[cogRY].JointValue(start, end);
		}

		// 1.2. precompute the trigo values
		//
		// 	a>0	 		    |				
		//      	        |				
		//	a=0		Z <-----+-----			Z' =  Z.cos - X.sin
		//     				|				X' =  X.cos + Z.sin
		//	a<0				|				
		//					v X				
		//
		cosine = cos(angle*3.141592654/180.0);
		sine = sin(angle*3.141592654/180.0);


//fprintf(stderr, "---------------- ORIENT %d --- [%d:%d] --------------\n",dancer,start,end);
//fprintf(stderr, "\nANGLE = %g\n",angle);
//fprintf(stderr, "\ncog x = "); for (ListIterator<Chunk> c(tX.data); c; ++c) {c()->Dump(TRUE); fprintf(stderr, "\n\t");}
//fprintf(stderr, "\ncog z = "); for (ListIterator<Chunk> c(tZ.data); c; ++c) {c()->Dump(TRUE); fprintf(stderr, "\n\t");}
//fprintf(stderr, "\n");

		// 2. rotate the cogX & cogZ
		if (angle != 0 && tX.IsUsed() && tZ.IsUsed()) {
//fprintf(stderr, "rotating %g   --   %g , %g  --  time = %d:%d\n",angle, cosine,sine,start,end);
			for(ListIterator<Chunk> cX(tX.data), cZ(tZ.data); cX && cZ; ++cX, ++cZ) {
				if (start<=cX()->startTime && cX()->endTime<=end) {
					tmp              =  cX()->startValue*cosine - cZ()->startValue*sine;
					cZ()->startValue =  cX()->startValue*sine   + cZ()->startValue*cosine;
					cX()->startValue =  tmp;

					tmp             =  cX()->leftValue*cosine - cZ()->leftValue*sine;
					cZ()->leftValue =  cX()->leftValue*sine   + cZ()->leftValue*cosine;
					cX()->leftValue =  tmp;

					tmp              =  cX()->rightValue*cosine - cZ()->rightValue*sine;
					cZ()->rightValue =  cX()->rightValue*sine   + cZ()->rightValue*cosine;
					cX()->rightValue =  tmp;

					tmp            =  cX()->endValue*cosine - cZ()->endValue*sine;
					cZ()->endValue =  cX()->endValue*sine   + cZ()->endValue*cosine;
					cX()->endValue =  tmp;
				}
			}
		}
	}

//fprintf(stderr, "\n--------------- tX & tZ   dancer %d , joint %d-%d , beats %d",dancer,start,end,nbBeats);
//fprintf(stderr, "\nCOG X = "); for (ListIterator<Chunk> c(tX.data); c; ++c) {c()->Dump(TRUE); fprintf(stderr, "\n\t");}
//fprintf(stderr, "\nCOG Z = "); for (ListIterator<Chunk> c(tZ.data); c; ++c) {c()->Dump(TRUE); fprintf(stderr, "\n\t");}


	if (start <0) {
		return;
	}

	if (dancers[dancer]->track[cogX].IsUsed())	tX.JointValue(start, end);
	if (dancers[dancer]->track[cogX].IsUsed())	tZ.JointValue(start, end);

//fprintf(stderr, "\n------------------\nCOG RY ="); for (ListIterator<Chunk> c(dancers[dancer]->track[cogRY].data); c; ++c) {c()->Dump(TRUE); fprintf(stderr, "\n\t");}
//fprintf(stderr, "\nCOG X = "); for (ListIterator<Chunk> c(tX.data); c; ++c) {c()->Dump(TRUE); fprintf(stderr, "\n\t");}
//fprintf(stderr, "\nCOG Z = "); for (ListIterator<Chunk> c(tZ.data); c; ++c) {c()->Dump(TRUE); fprintf(stderr, "\n\t");}
//fprintf(stderr, "\n");


	if (start == 0) {
//fprintf(stderr, "setting init values\n");
		if (dancers[dancer]->track[cogX].IsUsed()
			&& dancerInitPos[dancer][0] != dancers[dancer]->track[cogX].data->first()->startValue)
			dancers[dancer]->track[cogX].TranslateValue(dancerInitPos[dancer][0]);

		if (dancers[dancer]->track[cogY].IsUsed()
			&& dancerInitPos[dancer][1] != dancers[dancer]->track[cogY].data->first()->startValue)
			dancers[dancer]->track[cogY].TranslateValue(dancerInitPos[dancer][1]);

		if (dancers[dancer]->track[cogZ].IsUsed()
			&& dancerInitPos[dancer][2] != dancers[dancer]->track[cogZ].data->first()->startValue)
			dancers[dancer]->track[cogZ].TranslateValue(dancerInitPos[dancer][2]);

		if (dancers[dancer]->track[cogRX].IsUsed()
		   && dancerInitPos[dancer][3] != dancers[dancer]->track[cogRX].data->first()->startValue)
			dancers[dancer]->track[cogRX].TranslateValue(dancerInitPos[dancer][3]);

		if (dancers[dancer]->track[cogRZ].IsUsed()
		   && dancerInitPos[dancer][5] != dancers[dancer]->track[cogRZ].data->first()->startValue)
			dancers[dancer]->track[cogRZ].TranslateValue(dancerInitPos[dancer][5]);
	}
}


//================================================================================================

void Routine::ModifyValue(int time, int dancer, int track, int cmd, int vx, double vy,
						  int *beforeTime, int *afterTime)
{
enum {point, handle, handlelock, handletime};
static ListIterator<Chunk> c;
static int lastNumber = -1, lastDancer = -1, lastTrack = -1;
class Modif modif;


	// optimization : check if we are still moving the same chunk
	if ((lastNumber == number) && (lastDancer == dancer) && (lastTrack == track) && c &&
		(time==c()->startTime || time==c()->startTime+1 
		 || time==c()->endTime-1 || time==c()->endTime)) {
//fprintf(stderr, "cached ModifyValue  ");c()->Dump(TRUE);fprintf(stderr, "\n");
	} else {
		c.reset(dancers[dancer]->track[track].data);
		lastNumber = number;
		lastDancer = dancer;
		lastTrack = track;
//fprintf(stderr, "reset cache         ");c()->Dump(TRUE);fprintf(stderr, "\n");
	}

	// if needed, search the chunk and modify it
	for(; c; ++c) {
		if (time == c()->startTime) {
			//---- START TIME ----
			modif.type = VAL_VALUE;
			modif.time = time;
			modif.data.value = vy;
			modif.oldData.value = c()->endValue;
			RegisterModif(dancer, track, &modif);
			c()->startValue = vy;
			c()->isSynthetic = FALSE;
			// set linear
			modif.type = VAL_LEFT;
			  modif.time = c()->startTime;
			  modif.oldData.leftValue = c()->leftValue;
			  modif.oldData.leftTime = c()->leftTime;
			  c()->SetLinearLeft();
			  modif.data.leftValue = c()->leftValue;
			  modif.data.leftTime = c()->leftTime;
			  RegisterModif(dancer, track, &modif);
			modif.type = VAL_RIGHT;
			  if (afterTime) *afterTime = c()->endTime;
			  modif.time = c()->endTime;
			  modif.oldData.rightValue = c()->rightValue;
			  modif.oldData.rightTime = c()->rightTime;
			  c()->SetLinearRight();
			  modif.data.rightValue = c()->rightValue;
			  modif.data.rightTime = c()->rightTime;
			  RegisterModif(dancer, track, &modif);

			// handle previous chunk
			if (!c.isHead()) {
				c.prev()->endValue = vy;
				// set linear
				modif.type = VAL_LEFT;
				  if (beforeTime) *beforeTime = c.prev()->startTime;
				  modif.time = c.prev()->startTime;
				  modif.oldData.leftValue = c.prev()->leftValue;
				  modif.oldData.leftTime = c.prev()->leftTime;
				  c.prev()->SetLinearLeft();
				  modif.data.leftValue = c.prev()->leftValue;
				  modif.data.leftTime = c.prev()->leftTime;
				  RegisterModif(dancer, track, &modif);
				modif.type = VAL_RIGHT;
				  modif.time = c.prev()->endTime;
				  modif.oldData.rightValue = c.prev()->rightValue;
				  modif.oldData.rightTime = c.prev()->rightTime;
				  c.prev()->SetLinearRight();
				  modif.data.rightValue = c.prev()->rightValue;
				  modif.data.rightTime = c.prev()->rightTime;
				  RegisterModif(dancer, track, &modif);
			}
			break;

		} else if (time == c()->endTime) {
			//---- END TIME ----
			modif.type = VAL_VALUE;
			modif.time = time;
			modif.data.value = vy;
			modif.oldData.value = c()->endValue;
			RegisterModif(dancer, track, &modif);
			c()->endValue = vy;
			c()->isSynthetic = FALSE;
			// set linear
			modif.type = VAL_LEFT;
			  if (beforeTime) *beforeTime = c()->startTime;
			  modif.time = c()->startTime;
			  modif.oldData.leftValue = c()->leftValue;
			  modif.oldData.leftTime = c()->leftTime;
			  c()->SetLinearLeft();
			  modif.data.leftValue = c()->leftValue;
			  modif.data.leftTime = c()->leftTime;
			  RegisterModif(dancer, track, &modif);
			modif.type = VAL_RIGHT;
			  modif.time = c()->endTime;
			  modif.oldData.rightValue = c()->rightValue;
			  modif.oldData.rightTime = c()->rightTime;
			  c()->SetLinearRight();
			  modif.data.rightValue = c()->rightValue;
			  modif.data.rightTime = c()->rightTime;
			  RegisterModif(dancer, track, &modif);

			if (!c.isTail()) {
				c.next()->startValue = vy;
				// if last chunk if synthetic and entTime=nbBeats, then is it a terminal
				// chunk : endValue must be equal to startValue, else remove the synthetic flag
				if (c.next()->endTime != nbBeats) {
					c.next()->isSynthetic = FALSE;
				} else {
					if (c.next()->isSynthetic) {
						c.next()->endValue = c.next()->startValue;
					}
				}
				// set linear
				modif.type = VAL_LEFT;
				  modif.time = c.next()->startTime;
				  modif.oldData.leftValue = c.next()->leftValue;
				  modif.oldData.leftTime = c.next()->leftTime;
				  c.next()->SetLinearLeft();
				  modif.data.leftValue = c.next()->leftValue;
				  modif.data.leftTime = c.next()->leftTime;
				  RegisterModif(dancer, track, &modif);
				modif.type = VAL_RIGHT;
				  if (afterTime) *afterTime = c.next()->endTime;
				  modif.time = c.next()->endTime;
				  modif.oldData.rightValue = c.next()->rightValue;
				  modif.oldData.rightTime = c.next()->rightTime;
				  c.next()->SetLinearRight();
				  modif.data.rightValue = c.next()->rightValue;
				  modif.data.rightTime = c.next()->rightTime;
				  RegisterModif(dancer, track, &modif);
			}
			break;

		} else if (time == c()->startTime+1) {
			//---- LEFT HANDLE ----
			if (vx<=c()->startTime) vx=c()->startTime+1;
			if (vx>=c()->endTime) vx=c()->endTime-1;

			modif.type = VAL_LEFT;
			modif.time = time-1;
			modif.oldData.leftValue = c()->leftValue;
			modif.oldData.leftTime = c()->leftTime;

			c()->leftValue = vy;
			if (cmd == handletime) {
				c()->leftTime = vx;
				if (c()->rightTime == -1) {
					c()->rightTime = c()->startTime + (c()->endTime-c()->startTime)*2/3;
				}
			}

			modif.data.leftValue = c()->leftValue;
			modif.data.leftTime = c()->leftTime;
			RegisterModif(dancer, track, &modif);

			if (!c.isHead()) {
				if (cmd == handlelock) {
					double t1, t2, v1;
					v1 = vy - c()->startValue;
					t1 = c()->leftTime==-1 ? (c()->endTime-c()->startTime)/3
										   : c()->leftTime-c()->startTime;
					t2 = c.prev()->rightTime==-1 ? (c.prev()->endTime-c.prev()->startTime)/3
										   : c.prev()->endTime-c.prev()->rightTime;

					modif.type = VAL_RIGHT;
					modif.oldData.rightValue = c.prev()->rightValue;
					modif.oldData.rightTime = c.prev()->rightTime;

					c.prev()->rightValue = c.prev()->endValue - t2/t1*v1;
									 
					modif.data.rightValue = c.prev()->rightValue;
					modif.data.rightTime = c.prev()->rightTime;
					RegisterModif(dancer, track, &modif);
				}
			}
			break;

		} else if (time == c()->endTime-1) {
			//---- RIGHT HANDLE ----
			if (vx<=c()->startTime) vx=c()->startTime+1;
			if (vx>=c()->endTime) vx=c()->endTime-1;

			modif.type = VAL_RIGHT;
			modif.time = time+1;
			modif.oldData.rightValue = c()->rightValue;
			modif.oldData.rightTime = c()->rightTime;

			c()->rightValue = vy;
			if (cmd == handletime) {
				c()->rightTime = vx;
				if (c()->leftTime == -1) {
					c()->leftTime = c()->startTime + (c()->endTime-c()->startTime)/3;
				}
			}

			modif.data.rightValue = c()->rightValue;
			modif.data.rightTime = c()->rightTime;
			RegisterModif(dancer, track, &modif);

			if (!c.isTail()) {
				if (cmd == handlelock) {
					double t1, t2, v1;
					v1 = vy - c()->endValue;
					t1 = c()->rightTime==-1 ? (c()->endTime-c()->startTime)/3
										   : c()->endTime-c()->rightTime;
					t2 = c.next()->rightTime==-1 ? (c.next()->endTime-c.next()->startTime)/3
										   : c.next()->leftTime-c.next()->startTime;

					modif.type = VAL_LEFT;
					modif.oldData.leftValue = c.next()->leftValue;
					modif.oldData.leftTime = c.next()->leftTime;

					c.next()->leftValue = c.next()->startValue - t2/t1*v1;

					modif.data.leftValue = c.next()->leftValue;
					modif.data.leftTime = c.next()->leftTime;
					RegisterModif(dancer, track, &modif);
				}
			}
			break;
		}
	}
}

//------------------------------------------------------------------------------------------------

void Routine::ModifyType(int time, int dancer, int track, int cmd)
{
class Modif m;
List<TimeRange> *times = ListKeysTime(dancer, track);
ListIterator<TimeRange> t(times);
ListIterator<Chunk> c(dancers[dancer]->track[track].data);


	// foreach chunk NOT IN a figure, try to apply command
	while(t) {
		// find start of time range in chunk
		while (c && c()->startTime < t()->start)
			++c;

		if (!c)
			break; // no data for the track in the time range

		// while in time range
		while (c && c()->startTime < t()->end && c()->endTime <= t()->end) {
			if (cmd == SMOOTH_ALL) {
				if (!c.isHead()) {
//fprintf(stderr, "smoothing  %d --- %d   ",c()->startTime != t()->start,c()->endTime != t()->end);c()->Dump(TRUE);fprintf(stderr, "\n");

					c()->Smooth(c.prev(), c()->startTime != t()->start,
										  c()->endTime != t()->end);
				}
			} else if (time == c()->startTime || cmd == LIN_ALL) {
//fprintf(stderr, "linear %d",cmd);c()->Dump(TRUE);fprintf(stderr, "\n");

				switch (cmd) {
					case LIN_TIME:	c()->SetLinearTimes();
									break;

					case LIN_VAL:	c()->SetLinearValues();
									break;

					case LIN_ALL:	c()->SetLinearTimes();
									c()->SetLinearValues();
									break;
				}
			}
		}

		// advance to next time range
		++t;
	}

	times->free();
	delete times;
}

//------------------------------------------------------------------------------------------------
// Returns TRUE if a new modif was created, FALSE otherwise

int Routine::RegisterModif(int dancer, int track, class Modif *modif)
{
int result, partial = FALSE;

//fprintf(stderr, "REGISTRING modif on [%d / %s] =", dancer, trackName[track]);modif->Dump();fprintf(stderr, "\n");
	for (ListIterator<Figure> f(figures); f; ++f) {
		result = f()->RegisterModif(dancer, track, modif);
//fprintf(stderr, "    figure %d:%d --> result = %d\n", f()->start, f()->end, result);
		if (result == 2 || result == -2) {
			partial = result;
		} else if (result == -1) {
			return TRUE;
		} else if (result == 1) {
			return FALSE;
		}
	}

	return (partial==-2 ? TRUE : FALSE);
}

//================================================================================================

int jointToTrack[][4] = {
	{torsoX, torsoY, torsoZ, -1},		/*  0 */
	{pelvisX, pelvisY, pelvisZ, -1},
	{-1, -1, -1, -1},
	{headX, headY, headZ, -1},
	{headX, headY, headZ, -1},

	{lHipX, lHipY, lHipZ, -1},			/*  5 */
	{lKneeX, -1, -1, -1},
	{lAnkleX, lAnkleY, lAnkleZ, -1},
	{lFootX, -1, -1, -1},
	{lShoulderX, lShoulderY, lShoulderZ, lShoulderTY},
	{lElbowX, lElbowY, -1, -1},			/* 10 */
	{lWristX, lWristY, lWristZ, -1},

	{rHipX, rHipY, rHipZ, -1},
	{rKneeX, -1, -1, -1},
	{rAnkleX, rAnkleY, rAnkleZ, -1},
	{rFootX, -1, -1, -1},				/* 15 */
	{rShoulderX, rShoulderY, rShoulderZ, rShoulderTY},
	{rElbowX, rElbowY, -1, -1},
	{rWristX, rWristY, rWristZ, -1},

	{cogX, cogZ, -1, -1},
	{cogY, -1, -1, -1},					/* 20 */
	{cogRX, -1, -1, -1},
	{cogRZ, -1, -1, -1},
	{cogRY, -1, -1, -1} };

int Routine::GetKeys(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const
{
int key, dancer = -1, start = 0, end = nbBeats;
Tcl_Obj *res, *list;

	if (objc < 4) {
		Tcl_WrongNumArgs(interp, 2, objv, "dancer joint ?start end?");
		return TCL_ERROR;
	}

	// get the args	
	Tcl_GetIntFromObj(interp, objv[2], &dancer);

	// analyze the joint
	if (Tcl_GetIndexFromObj(interp, objv[3], jointName, "joint", 0, &key) != TCL_OK) {
		return TCL_ERROR;
	}

	// check dancer number
	if (dancer == -1 && (key >= NB_JOINTS)) {
		dancer = COMMENTS;
	} else if (dancer<0 || dancer>=nbDancers) {
		Tcl_SetResult(interp, "bad dancer index", 0);
		return TCL_ERROR;
	}

	// get the args	
	if (objc > 4) Tcl_GetIntFromObj(interp, objv[4], &start);
	if (objc > 5) Tcl_GetIntFromObj(interp, objv[5], &end);

	// do it
	res = Tcl_NewListObj(0, NULL);
	if (key >= NB_JOINTS) {
		// special case for notes
		for(ListIterator<Note> n(notes[dancer][key == NB_JOINTS ? REAL : INCLUDED]); n; ++n) {
			list = Tcl_NewListObj(0, NULL);
			Tcl_ListObjAppendElement(interp, list, Tcl_NewIntObj(n()->start()));
			Tcl_ListObjAppendElement(interp, list, Tcl_NewIntObj(n()->end()));
			Tcl_ListObjAppendElement(interp, list, Tcl_NewIntObj(n()->number()));

			Tcl_ListObjAppendElement(interp, res, list);
		}		
	} else {
		AddKeys(interp, res, dancer, jointToTrack[key][0]);
		AddKeys(interp, res, dancer, jointToTrack[key][1]);
		AddKeys(interp, res, dancer, jointToTrack[key][2]);
		AddKeys(interp, res, dancer, jointToTrack[key][3]);

		// we want unique time index : remove the duplicates & values not in [start:end]
		int len, value, value2;
		Tcl_Obj *elem;
		Tcl_ListObjLength(interp, res, &len);
		for (int i=0; i<len; i++) {
			Tcl_ListObjIndex(interp, res, i, &elem);
			Tcl_GetIntFromObj(interp, elem, &value);
			for (int j=i+1; j<len; j++) {
				Tcl_ListObjIndex(interp, res, j, &elem);
				Tcl_GetIntFromObj(interp, elem, &value2);
				if (value2 == value || abs(value2) < start || abs(value2) > end) {
					Tcl_ListObjReplace(interp, res, j, 1, 0, NULL);
					len--;
					j--;
				}
			}
		}
	}

	// return the list of keys
	Tcl_SetObjResult(interp, res);
	return TCL_OK;
}

void Routine::AddKeys(Tcl_Interp *interp, Tcl_Obj *res, int dancer, int track) const
{
List<TimeRange> *timeRange;
ListIterator<Chunk> c;
ListIterator<TimeRange> t;


	if (track == -1 || !dancers[dancer]->track[track].IsUsed())
		return;

	// get the keys in track
	timeRange = ListKeysTime(dancer, track);

//fprintf(stderr, "\nDancer %d/%-16s = ", dancer, trackName[track]); for(ListIterator<TimeRange> tt(timeRange); tt; ++tt) { fprintf(stderr, "%d-%d  ", tt()->start, tt()->end); }

	t.reset(timeRange);

	for(c.reset(dancers[dancer]->track[track].data); c; ++c) {
		// if synthetic chunk, skip it
//		if (c()->isSynthetic) {
//			continue;
//		}

		// check if key in routine or not
		if (t && (c()->startTime > t()->start || t()->start==0)) {
//fprintf(stderr, " %d,", c()->startTime);
			Tcl_ListObjAppendElement(interp, res, Tcl_NewIntObj(c()->startTime));
		} else {
//fprintf(stderr, " -%d,", c()->startTime);
			Tcl_ListObjAppendElement(interp, res, Tcl_NewIntObj(c()->startTime == 0 
																? -1 : -c()->startTime));
		}

		// if tail, handle last end time
		if (c.isTail()) {
//fprintf(stderr, "\nTail  ");
			if (t && c()->endTime >= t()->end) {
				if (!c()->isSynthetic) {
//fprintf(stderr, " !%d,", c()->endTime);
					Tcl_ListObjAppendElement(interp, res, Tcl_NewIntObj(c()->endTime));
				}
			} else {
//fprintf(stderr, " !-%d,", c()->endTime);
				Tcl_ListObjAppendElement(interp, res, Tcl_NewIntObj(c()->endTime == 0 
																	? -1 : -c()->endTime));
			}
		}

		// if needed, advance to the key range
		if (t && c()->endTime >= t()->end)
			++t;
	}

	// get the MODIFs in FIGURE (set them as minus time for distinction)
	for(ListIterator<Figure> f(figures); f; ++f) {
		for(int j=0; j<f()->nbMap; j++) {
			if (f()->map[j] == dancer) {
				if (f()->flags[j][track] & USED) {
					// modification of figure
					for(ListIterator<Modif> m(f()->modifications[j][track]); m; ++m) {
//fprintf(stderr, " <%d>,", m()->time);
						if (m()->type & VAL_VALUE) {
							Tcl_ListObjAppendElement(interp, res, Tcl_NewIntObj(m()->time));
						}
					}
					break;			
				}
			}
		}
	}

	// free the resources
	timeRange->free();
	delete timeRange;
}

//------------------------------------------------------------------------------------------------

int Routine::SearchConflict(Routine &source, int *min, int *max, int srcDancer, int destDancer,
							int joint, int timeInsert)
{
	for(int i=0; i<4 && jointToTrack[joint][i]!=-1; i++) {
		// if track not used (either source or dest), continue
		if (!dancers[destDancer]->track[jointToTrack[joint][i]].IsUsed())
			continue;
		if (!source.dancers[srcDancer]->track[jointToTrack[joint][i]].IsUsed())
			continue;

		// adjust the begin of the tracks
		ListIterator<Chunk> srcC(source.dancers[srcDancer]->track[jointToTrack[joint][i]].data);
		ListIterator<Chunk> destC(dancers[destDancer]->track[jointToTrack[joint][i]].data);

//fprintf(stderr, "\n---- time insert = %d    --   track = %s", timeInsert, trackName[jointToTrack[joint][i]]);
		while (destC && destC()->startTime < timeInsert) {
			++destC;
		}
//fprintf(stderr, "\nsrc = ");for (ListIterator<Chunk> __c(srcC); __c; ++__c) {__c()->Dump(); if (__c.isTail()) __c()->DumpEnd();}
//fprintf(stderr, "\ndest = ");for (ListIterator<Chunk> __c(destC); __c; ++__c) {__c()->Dump(); if (__c.isTail()) __c()->DumpEnd();}

		
		// max time = timeInsert + source.nbBeats
		while (destC && destC()->endTime <= timeInsert+source.nbBeats) {
			if (!destC()->isSynthetic) {
//fprintf(stderr,"\n=== Conflict at "); destC()->Dump(TRUE);
				return TRUE;
			}
			++destC;
		}

// routine; routine1 load Waltz/Basis/waltz1; routine0 checkinsertion routine1 3000 {-1 1}

// routine0 insert routine1 "toto" pink true false true 3000 {-1 1}

// routine; routine2 load Waltz/Basis/waltz2; routine0 insert routine2 "toto" red true false true 6000 {-1 1}
	}

//fprintf(stderr,"\n=== OK   %s", jointName[joint]);
	return FALSE;
}

//------------------------------------------------------------------------------------------------

int Routine::Delete(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
int joint, dancer = -1, start = -1, end = nbBeats, j;
Tcl_Obj *res, *list;


	if (objc < 4) {
		Tcl_WrongNumArgs(interp, 2, objv, "dancer joint ?start end?");
		return TCL_ERROR;
	}

	// get the args	
	Tcl_GetIntFromObj(interp, objv[2], &dancer);
	if (dancer<0 || dancer>=nbDancers) {
		Tcl_SetResult(interp, "bad dancer index", 0);
		return TCL_ERROR;
	}
	// analyze the joint
	if (Tcl_GetIndexFromObj(interp, objv[3], jointName, "joint", 0, &joint) != TCL_OK) {
		return TCL_ERROR;
	}
	if (joint >= NB_JOINTS) {
		Tcl_SetResult(interp, "bad joint", 0);
		return TCL_ERROR;
	}
	// get start and end time
	if (objc > 4) Tcl_GetIntFromObj(interp, objv[4], &start);
	if (objc > 5) Tcl_GetIntFromObj(interp, objv[5], &end);

	// do it
	if (start == -1 && end == -1) {
		//-----------------------
		// remove the whole track
		for(j=0; j<4 && jointToTrack[joint][j]!=-1; j++) {
//fprintf(stderr, "DELETE WHOLE RANGE on [%d - %s]\n", dancer,trackName[jointToTrack[joint][j]]);
			dancers[dancer]->track[jointToTrack[joint][j]].Delete(-1,-1,-1);
		}
		// figure rejection
		for(ListIterator<Figure> f(figures); f; ++f) {
			for(int i=0; i<f()->nbMap; i++) {
				// search if dancer used in figure
				if (f()->map[i] == dancer) {
					for(j=0; j<4 && jointToTrack[joint][j]!=-1; j++) {
						// foreach used joint, reject it
						if (f()->flags[i][jointToTrack[joint][j]] & USED) {
							f()->flags[i][jointToTrack[joint][j]] &= ~USED;
							f()->flags[i][jointToTrack[joint][j]] |= REJECT;
						}
					}
				}
			}
		}

	} else {
		//-------------------------------
		// first check if range in figure
		for(j=0; j<4 && jointToTrack[joint][j]!=-1; j++) {
			for(ListIterator<Figure> f(figures); f; ++f) {
				// remove modif
				if (f()->DeleteModif(*this, dancer, jointToTrack[joint][j], start, end)) {
					break;
				}
			}
		}

		//-------------------------
		// remove the keys in track
		for(j=0; j<4 && jointToTrack[joint][j]!=-1; j++) {
//fprintf(stderr, "GET key range for %d - %s\n", dancer,trackName[jointToTrack[joint][j]]);
			List<TimeRange> *times = ListKeysTime(dancer, jointToTrack[joint][j]);
			ListIterator<TimeRange> t(times);
			for(; t; ++t) {
//fprintf(stderr, "PROCESSING key range   %d:%d\n", t()->start,t()->end);
				int min = start<t()->start+1 ? t()->start+1 : start;
				int max = end>t()->end-1 ? t()->end-1 : end;
//fprintf(stderr, "Track::Delete  %s -- %d:%d, real %d:%d\n", trackName[jointToTrack[joint][j]], start, end, min, max);
				dancers[dancer]->track[jointToTrack[joint][j]].Delete(min,max, nbBeats);
				if (bodies[dancer]) {
//fprintf(stderr, "    Resync body at %d\n", time);
					dancers[dancer]->SetBody(bodies[dancer], time, jointToTrack[joint][j]);
				}
			}
			times->free();
			delete times;
		}
	}

	return TCL_OK;
}

//------------------------------------------------------------------------------------------------

int Routine::GetFigures(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const
{
int key, dancer = -1;
Tcl_Obj *res, *list;
ListIterator<Figure> f;


	if (objc !=2 && objc != 4) {
		Tcl_WrongNumArgs(interp, 2, objv, "?dancer joint?");
		return TCL_ERROR;
	}

	// if no args, list of filename
	if (objc == 2) {
		res = Tcl_NewListObj(0, NULL);
		for(f.reset(figures); f; ++f) {
			Tcl_ListObjAppendElement(interp, res, Tcl_NewStringObj(f()->filename, -1));
		}
		goto done;
	}

	// get the args	
	Tcl_GetIntFromObj(interp, objv[2], &dancer);
	if (dancer<0 || dancer>=nbDancers) {
		Tcl_SetResult(interp, "bad dancer index", 0);
		return TCL_ERROR;
	}

	// analyze the joint
	if (Tcl_GetIndexFromObj(interp, objv[3], jointName, "joint", 0, &key) != TCL_OK) {
		return TCL_ERROR;
	}

	res = Tcl_NewListObj(0, NULL);
	for(f.reset(figures); f; ++f) {
		for(int j=0; j<f()->nbMap; j++) {
			if (f()->map[j] == dancer) {
				if ((jointToTrack[key][0] != -1 && (f()->flags[j][jointToTrack[key][0]] & USED))
					|| (jointToTrack[key][1] != -1
						&& (f()->flags[j][jointToTrack[key][1]] & USED))
					|| (jointToTrack[key][2] != -1
						&& (f()->flags[j][jointToTrack[key][2]] & USED))
					|| (jointToTrack[key][3] != -1 
						&& (f()->flags[j][jointToTrack[key][3]] & USED)) ) {

					list = Tcl_NewListObj(0, NULL);
					Tcl_ListObjAppendElement(interp, list, Tcl_NewIntObj(f()->number));
					Tcl_ListObjAppendElement(interp, list, Tcl_NewIntObj(f()->start));
					Tcl_ListObjAppendElement(interp, list, Tcl_NewIntObj(f()->end));
					
					Tcl_ListObjAppendElement(interp, res, list);
				}
			}
		}
	}

	// return the list of routines
done:
	Tcl_SetObjResult(interp, res);
	return TCL_OK;
}

//================================================================================================

List<TimeRange> * Routine::ListKeysTime(int dancer, int track) const
{
int timeStart, startFigure, endFigure, oldStart;
List<TimeRange> *range = new List<TimeRange>;


	timeStart = 0;
	endFigure = -1;
	oldStart = -1;
	while (timeStart != nbBeats) {
		startFigure = 1000000000;
		for(ListIterator<Figure> f(figures); f; ++f) {
			if (f()->start < startFigure && f()->start > oldStart) {
				for(int j=0; j<f()->nbMap; j++) {
					if (f()->map[j] == dancer && (f()->flags[dancer][track] & USED)
						&& !(f()->flags[dancer][track] & REJECT)) {

						startFigure = f()->start;
						endFigure = f()->end;
						break;
					}
				}
			}
		}
		if (startFigure != 1000000000) {
			if (startFigure != 0 && startFigure != timeStart) {
				// TODO: verify +1 & -1
				range->append(new TimeRange(timeStart, startFigure));
			}
			timeStart = endFigure;
			oldStart = startFigure;
		} else {
			range->append(new TimeRange(timeStart, nbBeats));
			timeStart = nbBeats;
		}
	}

	return range;
}


//================================================================================================
//
//	SynthetizeNotes
//
//		Merge the notes contained in "real" with thoses in "included" in order to produce
//		a unique "real" notes track. All the "real" notes dominates the included "one"
//		which are more or less default notes.
//
//================================================================================================

void Routine::SynthetizeNotes(int timeOffset)
{
int i, start, end;


	// 1. for each dancer in the routine
	for(i=0; i<nbDancers; i++) {
		// 2. for each included note
		for(ListIterator<Note> ni(notes[i][INCLUDED]); ni; ++ni) {
			start = ni()->start();
			end = ni()->end();
			// 2.1. find if overlapping an other "real" note
			for(ListIterator<Note> nr(notes[i][REAL]); nr; ++nr) {
				if (start < nr()->start() && nr()->start() < end) {
					end = nr()->start();
				}
				if (start < nr()->end() && nr()->end() < end) {
					start = nr()->end();
				}
			}

			// 2.2. if no real notes exist, then create the track
			if (notes[i][REAL] == 0) {
				notes[i][REAL] = new List<Note>;
			}
			// 2.3. add the note which now doesn't overlap any other note
			notes[i][REAL]->append(new Note(start, end, ni()->text()));
			ni()->unlinkText(); // to avoid deletion of text
		}

		// 3. finally, offset time of each note
		for(ListIterator<Note> n(notes[i][REAL]); n; ++n) {
			n()->offset(timeOffset);
		}
	}
}
