#ifndef INCLUDED_ROUTINE
#define INCLUDED_ROUTINE

#include "global.h"

template<class T> class List;
class Chunk;
class Movie;


//----------------------------------------------------------------------------------------------- 
//	TimeRange
//------------------------------------------------------------------------------------------------

struct TimeRange {
	// DATA
	int start, end;

	// CREATORS
	TimeRange(int s, int e) : start(s), end(e) {}
	~TimeRange() {}
};

//----------------------------------------------------------------------------------------------- 
//	Modif
//------------------------------------------------------------------------------------------------

class Modif {
	// DATA
	int time;
	int type;
	struct {
		double value;
		double rightValue, leftValue;
		int rightTime, leftTime;
	} data, oldData;

	// FRIENDS
	friend class Routine;
	friend class Figure;
	friend class Track;
	friend int yyparse();

public:
	// CREATORS
	Modif();
	Modif(const Modif *modif);
	Modif(int flags, int t, double v=0, double rv=0, int rt=0, double lv=0, int lt=0);
	~Modif();

	// ACCESSORS
	void Save(FILE *out) const;
#ifdef DEBUG
	void Dump() const;
#endif
};


//----------------------------------------------------------------------------------------------- 
//	Key
//------------------------------------------------------------------------------------------------

class Key {
	// DATA
	int start, end;
	int dancer;
	class List<Chunk> *part[NB_DOF];

	// FRIENDS
	friend class Routine;
	friend int yyparse();

public:
	// CREATORS
	Key();
	~Key();

	// MANIPULATORS
	int Compile(class Interpolator *dancers[]);
	int CheckIntegrity();

	// ACCESSORS
#ifdef DEBUG
	void Dump() const;
#endif
};


//----------------------------------------------------------------------------------------------- 
//	Figure
//------------------------------------------------------------------------------------------------

#define REJECT	0x1
#define USED	0x2

class Figure {
	// DATA
	static int s_serial;
	int number;
	int start, end;
	char *filename;
	char *color;
	int nbMap;
	int map[MAX_DANCERS];
	class List<Modif> *modifications[MAX_DANCERS][NB_DOF];
	int flags[MAX_DANCERS][NB_DOF];
	char *name;
	int includeNotes;

	// FRIENDS
	friend class ViewTrack;
	friend class Routine;
	friend int yyparse();

public:
	// CREATORS
	Figure();
	~Figure();

	// MANIPULATORS
	int CheckIntegrity();
	int Compile(class Routine &dest);
	int DeleteModif(class Routine &routine, int dancer, int track, int startTime, int endTime);
	int Insert(class Routine &dest, class Routine &source);
	int RegisterModif(int dancer, int track, class Modif *modif);

	// ACCESSORS
	int ApplyToDancer(int dancer) const;
	void Save(FILE *out) const;
#ifdef DEBUG
	void Dump() const;
#endif
};


//----------------------------------------------------------------------------------------------- 
//	Note
//------------------------------------------------------------------------------------------------

class Note {
	// DATA
	static int s_serial;
	int d_number;
	int d_start, d_end;
	char *d_text;

public:
	// CREATORS
	Note(int start, int end, char *text) : d_number(s_serial++),
			d_start(start), d_end(end), d_text(text) {}
	~Note() { delete[] d_text; }

	// MANIPULATOR
	void modify(int start, int end, char *text) {
		d_end = end!=-1 ? end : start==-1 ? d_end : start + d_end-d_start;
		if (start != -1) d_start = start;
		if (text) {
			delete[] d_text;
			d_text = new char[strlen(text)+1];
			strcpy(d_text, text);
		}
	}
	void offset(int time) {
		d_start += time;
		d_end += time;
	}
	void unlinkText() { d_text = 0; }

	// ACCESSORS
	int number() const { return d_number; }
	char *text() const { return d_text; }
	int start() const { return d_start; }
	int end() const { return d_end; }

#ifdef DEBUG
	void Dump() const { if (d_start==d_end)
							fprintf(stderr, "\"%s\"@%d", d_text, d_start);
						else
							fprintf(stderr, "\"%s\"@[%d:%d]", d_text, d_start, d_end); }
#endif
};


//------------------------------------------------------------------------------------------------
//	A routine = dancers + high level description + interpolator ("compiled" version)
//------------------------------------------------------------------------------------------------

class Routine {
	// DATA
	int number;

	int nbDancers;
	class Interpolator *dancers[MAX_DANCERS];
	double dancerInitPos[MAX_DANCERS][6];
	char *dancerNames[MAX_DANCERS];
	int dancerSex[MAX_DANCERS];
	char *dancerTypes[MAX_DANCERS];
	class Body *bodies[MAX_DANCERS];

	int nbBeats, time, rhythm, bpm;
	char *description, *name, *copyright, *text;

	class List<Key> *keys;
	class List<Figure> *figures;
	class List<Movie> *movies;
	class List<Note> *notes[MAX_DANCERS+1][2];		// last one for globals notes in routine
	enum {REAL = 0, INCLUDED = 1, COMMENTS = MAX_DANCERS};

	// FRIENDS
	friend class ViewTrack;
	friend class ViewMovie;		// for dumping the list of movies
	friend class Figure;
	friend class Movie;
	friend int yyparse();

public:
	// CREATORS
	Routine(int counter);
	~Routine();

	// MANIPULATORS
	void ModifyValue(int time, int dancer, int track, int cmd, int vx, double vy,
					 int *beforeTime = 0, int *afterTime = 0);
	void ModifyType(int time, int dancer, int track, int cmd);
	int RegisterModif(int dancer, int track, class Modif *modif);

	int CheckInsertion(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Delete(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int HandleDancer(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int HandleNote(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int HandleMovie(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Insert(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Load(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Save(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Set(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Sync(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
	int Update(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);

	// ACCESSORS
	int ApplyTo(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const;
	int Get(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const;
	int GetKeys(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const;
	int GetFigureInfo(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const;
	int GetFigures(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) const;
	int Number() { return number; }

#ifdef DEBUG
	void Dump() const;
#endif

private:
	void AddKeys(Tcl_Interp *interp, Tcl_Obj *res, int dancer, int track) const;
	int Compile(char *filename, int mode);
	int Generate();
	void NewNbBeats();
	void OrientDancer(int dancer, int start, int end);
	int SearchConflict(Routine &source, int *min, int *max, int srcDancer, int destDancer,
					   int joint, int timeInsert);
	void SynthetizeNotes(int timeOffset);
	List<TimeRange> *ListKeysTime(int dancer, int track) const;
};

#endif
