#include <math.h>
#include "togl.h"

#include "list.h"
#include "track.h"
#include "routine.h"
#include "body.h"		// just for body->joints

extern char *trackName[];
extern Tcl_Interp *currentInterp;

//------------------------------------------------------------------------------------------------
//	Chunk
//------------------------------------------------------------------------------------------------

Chunk::Chunk(int start, int end, double val0, double val1, 
			 double left, double right, int tLeft, int tRight, int synthetic) :
	startTime(start), endTime(end), leftTime(tLeft), rightTime(tRight), 
	startValue(val0), endValue(val1), leftValue(left), rightValue(right),
	isSynthetic(synthetic)
{
}

//------------------------------------------------------------------------------------------------

void Chunk::SetLinear()
{
	leftValue = startValue + (endValue-startValue)/3;
	rightValue = startValue + (endValue-startValue)*2/3;
	leftTime = rightTime = -1;
}

void Chunk::SetLinearValues()
{
	leftValue = startValue + (endValue-startValue)/3;
	rightValue = startValue + (endValue-startValue)*2/3;
}

void Chunk::SetLinearTimes()
{
	leftTime = rightTime = -1;
}

void Chunk::SetLinearLeft()
{
	leftValue = startValue + (endValue-startValue)/3;
	leftTime = -1;
}

void Chunk::SetLinearRight()
{
	rightValue = startValue + (endValue-startValue)*2/3;
	rightTime = -1;
}

void Chunk::Smooth(class Chunk *prev, int modifyPrev, int modifyCurrent)
{
double ldx, ldy, rdx, rdy, ltheta, rtheta, theta, ldist, rdist;

	
	ldx = startTime - prev->startTime;
	ldy = startValue - prev->startValue;
	ldist = sqrt(ldx*ldx+ldy*ldy);
	ltheta = atan2(ldy, ldx+0.00000001);

	rdx = endTime - startTime;
	rdy = endValue - startValue;
	rdist = sqrt(rdx*rdx+rdy*rdy);
	rtheta = atan2(rdy, rdx+0.00000001);

	theta = (ltheta/ldist + rtheta/rdist)/(1/ldist+1/rdist);

	if (modifyPrev) {
		if (prev->rightTime == -1) {
			prev->rightValue = prev->endValue - tan(theta)*(prev->endTime-prev->startTime)/3;
		} else {
			prev->rightTime = (int) (prev->endTime - ldist*cos(theta));
			prev->rightValue = prev->endValue - ldist*sin(theta);
		}
	}

	if (modifyCurrent) {
		if (leftTime == -1) {
			leftValue = startValue + tan(theta)*(endTime-startTime)/3;
		} else {
			leftTime = (int) (startTime + rdist*cos(theta));
			leftValue = startValue + rdist*sin(theta);
		}
	}
}

//------------------------------------------------------------------------------------------------

double Chunk::operator[](int time) const
{
register double t, t2, t3, u, u2, u3;

	if (endTime == startTime) {
		return startValue;
	}
	t = (time-startTime)/((double) (endTime-startTime));

	if (leftTime != -1 || rightTime != -1) {
		t2 = t*t;
		t3 = t2*t;
		u = 1.0 - t;
		u2 = u*u;
		u3 = u2*u;
		t = startTime*u3 + 3.0*(leftTime*t*u2 + rightTime*t2*u) + endTime*t3;
		t = (t-startTime)/((double) (endTime-startTime));
	}

	t2 = t*t;
	t3 = t2*t;
	u = 1.0 - t;
	u2 = u*u;
	u3 = u2*u;

	return (startValue*u3 + 3.0*(leftValue*t*u2 + rightValue*t2*u) + endValue*t3);
}

//------------------------------------------------------------------------------------------------

void Chunk::Draw() const
{
	int step = (endTime-startTime)/20;
	if (step == 0) {
		return;
	}
	for(int time = startTime; time <= endTime; time += step) {
		glVertex2d(time, (*this)[time]);
	}
}

void Chunk::DrawHandle() const
{
	glLoadName(startTime+1);	// time
	glBegin(GL_LINES);
	  glVertex2d(startTime, startValue);
	  glVertex2d(leftTime==-1 ? startTime+(endTime-startTime)/3 : leftTime, leftValue);
	glEnd();
	glBegin(GL_POINTS);
	  glVertex2d(leftTime==-1 ? startTime+(endTime-startTime)/3 : leftTime, leftValue);
	glEnd();

	glLoadName(endTime-1);	// time
	glBegin(GL_LINES);
	  glVertex2d(rightTime==-1 ? startTime+(endTime-startTime)*2/3 : rightTime, rightValue);
	  glVertex2d(endTime, endValue);
	glEnd();
	glBegin(GL_POINTS);
	  glVertex2d(rightTime==-1 ? startTime+(endTime-startTime)*2/3 : rightTime, rightValue);
	glEnd();
}

//------------------------------------------------------------------------------------------------

void Chunk::Save(FILE *out) const
{
double left = startValue + (endValue-startValue)/3;
double right = startValue + (endValue-startValue)*2/3;


	if (isSynthetic) {
		return;
	}

	if (fabs(left-leftValue) < 0.000001 && fabs(right-rightValue) < 0.000001
		&& leftTime == -1 && rightTime == -1) {
		fprintf(out, " %g@%d linear", startValue, startTime);
	} else {
		fprintf(out, " %g@%d bezier(%g,%g, %d,%d)", startValue, startTime,
				leftValue, rightValue, leftTime, rightTime);
	}
}



//================================================================================================
//
//		A track is a collection of Chunk representing pieces of interpolation
//
//================================================================================================

int Track::FindPlace(int time)
{
int end = 1000000000;
class Chunk *c;


//fprintf(stderr, "\nsearching for %d  at ", time); current()->Dump(TRUE);

	// if inside the current chunk
	if (current()->startTime <= time && time < current()->endTime) {
		;
	} else if (time < current()->startTime) {
		// find the right chunk, before the current one
		while (current()->startTime > time && !current.isHead()) {
			--current;
		}
	} else {
		// find the right chunk, after the current one
		while (current()->endTime <= time && !current.isTail()) {
			++current;
		}
	}

	if (!current.isTail()) {
		end = current.next()->startTime;
	}

	return end;
}

int Track::AddInPlace(class Chunk *c)
{
int end;

//fprintf(stderr, "add in place  %d:%d\n", c->start, c->end);

	if (!current) {
		data->append(c);
		current.reset(data);
		return TRUE;
	}

	end = FindPlace(c->startTime);

	while ((current()->startTime <= c->startTime && c->startTime < current()->endTime)
		|| (c->endTime > end)) {
		// conflict, check if we can remove synthetic chunk
		if (current()->isSynthetic) {
//fprintf(stderr, "\n          REMOVING  ");current()->Dump(TRUE);
			data->remove(current());
			current.reset(data);
		} else {
#ifndef WIN32
//fprintf(stderr, "\n<chunk> overwriting part in time range %d-%d", current()->startTime, current()->endTime); current()->Dump(TRUE);
//fprintf(stderr, "\n");for (ListIterator<Chunk> __c(data); __c; ++__c) __c()->Dump(TRUE); data->last()->DumpEnd(); fprintf(stderr, "\n");
#endif

			sprintf(error, "overwriting part in time range %d-%d in ",
					current()->startTime, current()->endTime);
			Tcl_AppendResult(currentInterp, error, NULL);
			return FALSE;
		}

		end = FindPlace(c->startTime);
	}

	if (current.isHead() && c->startTime < current()->startTime) {
//fprintf(stderr, "\n          PREPEND "); c->Dump(TRUE);
		data->prepend(c);
		current.reset(data);
	} else {
		if (c->startTime < current()->startTime) {
			--current;
		}
//fprintf(stderr, "\n          INSERT @ %d = ",current()->startTime); c->Dump(TRUE);
		data->insert(current, c);
		++current;
	}
	return TRUE;
}


int Track::AddInPlace(class Track *t)
{
class Chunk *c = t->data->first(), *last = t->data->last();
int end;


	if (!current) {
		// swap the lists
		data = t->data;
		t->data = NULL;
		current.reset(data);
		return TRUE;
	}

	end = FindPlace(c->startTime);

//if (end != current()->endTime) fprintf(stderr, "\n!!!!!!!!!!!!!!!!!!! end != current->endTime");
//fprintf(stderr, "\nend=%10d -- ",end);c->Dump(TRUE);fprintf(stderr, "\n               -- ");current()->Dump(TRUE);

	while ((current()->startTime <= c->startTime && c->startTime < current()->endTime)
		|| (c->startTime < current()->endTime && last->endTime > current()->startTime)
		|| (last->endTime > end)) {
		// conflict, check if we can remove synthetic chunk
		if (current()->isSynthetic) {
//fprintf(stderr, "\nremoving ");current()->Dump(TRUE);
			data->remove(current());
			current.reset(data);
		} else {
#ifndef WIN32
//fprintf(stderr, "\n<track> overwriting part in time range %d-%d", current()->startTime, current()->endTime); current()->Dump(TRUE);
//fprintf(stderr, "\n");for (ListIterator<Chunk> __c(data); __c; ++__c) __c()->Dump(TRUE); data->last()->DumpEnd(); fprintf(stderr, "\n");
#endif

			sprintf(error, "overwriting part in time range %d-%d in ",
					current()->startTime, current()->endTime);
			Tcl_AppendResult(currentInterp, error, NULL);
			return FALSE;
		}

		end = FindPlace(c->startTime);
//if (end != current()->endTime) fprintf(stderr, "\n!!!!!!!!!!!!!!!!!!! end != current->endTime");
//fprintf(stderr, "\nEND=%10d -- ",end);c->Dump(TRUE);fprintf(stderr, "\n               -- ");current()->Dump(TRUE);
	}

	// do the insertion
	if (current.isHead() && c->startTime < current()->startTime) {
		data->prepend(t->data);
		current.reset(data);
	} else {
		if (c->startTime < current()->startTime) {
			--current;
		}
		data->insert(current, t->data);
		++current;
	}

	// disconnect the chunks in the old track after putting them in the new track
	// (avoid the deletion of the objects)
	t->data = NULL;

	return TRUE;
}

//------------------------------------------------------------------------------------------------

void Track::Delete(int start, int end, int nbBeats)
{
	if (start== -1) {
		// free the whole track
		data->free();
		delete data;
		data = new List<Chunk>;

	} else if (start == nbBeats) {
		if (current) {
			Chunk *c = data->last();
			if (c->endTime == nbBeats) {
				c->endValue = c->startValue;
				c->SetLinear();
				c->isSynthetic = TRUE;
			}
		}

	} else {
		// delete the keys between start & end
		ListIterator<Chunk> c(data);
		int needNew = FALSE;

		// 1. if starting if first chunk
		if (c && c()->startTime<=start && end<c()->endTime) {
//fprintf(stderr, "    proc. ");c()->Dump(TRUE);fprintf(stderr, "\n");
			// 1.1. while in the range delete the chunk
			while (c && (   (c()->startTime<=start && c()->endTime<end)
					     || (start==end && c()->startTime==start))) {
//fprintf(stderr, "        DESTROYING\n");
				data->destroy(c());
				needNew = TRUE;
				c.reset(data);
//fprintf(stderr, "    +proc ");if(c) {c()->Dump(TRUE);}else{fprintf(stderr, "(empty)");} fprintf(stderr, "\n");
			}
			// 1.2. create a new chunk for filling the gap between 0-t
			if (c && needNew) {
//fprintf(stderr, "        NEW PREPEND\n");
				data->prepend(new Chunk(0, c()->startTime, 0,c()->startValue, 
										c()->startValue/3,c()->startValue*2/3, -1,-1, TRUE));
			}

		} else if (c) {
			// 2. delete the range inside the track
//fprintf(stderr, "    --- skipping --- ");c()->Dump(TRUE);fprintf(stderr, "\n");
			// 2.1. find the first chunk targeted by the deletion
			while (!c.isTail() && c()->endTime<start) {
				++c;
			}

//fprintf(stderr, "    real. ");c()->Dump(TRUE);fprintf(stderr, "\n");
//fprintf(stderr, "    proc. ");if(!c.isTail()) {c.next()->Dump(TRUE);}else{fprintf(stderr, "(empty)");} fprintf(stderr, "\n");
			while (!c.isTail() && c.next()->startTime<=end) {
				// 2.2. while in the range delete the chunks
//fprintf(stderr, "        DESTROYING\n");
				data->destroy(c.next());
//fprintf(stderr, "    +proc ");if(!c.isTail()) {c.next()->Dump(TRUE);}else{fprintf(stderr, "(empty)");} fprintf(stderr, "\n");
			}

			if (c.isTail()) {
				// 2.3.1. set a proper ending for the track
				c()->endTime = nbBeats;
				c()->endValue = c()->startValue;
				c()->SetLinear();
//				if (c.isHead() && end == nbBeats) {
				c()->isSynthetic = TRUE;
//				}
			} else {
				// 2.3.2. create a new chunk for filling the gap
				c()->endTime = c.next()->startTime;
				c()->endValue = c.next()->startValue;
			}
			// 2.4. adjust remaing chunk
			c()->SetLinear();
			if (start != end) {
				c()->isSynthetic = TRUE;
			}
		}
	}

	// reset current (we could have destroyed some elements)
	current.reset(data);
}

//------------------------------------------------------------------------------------------------

double Track::operator[](int time)
{
	if (!current) {
fprintf(stderr, "(ERROR) Track::operator[] : current is null\n");
		return 0.0;
	}

	// if inside the current chunk
	if (current()->startTime <= time && time < current()->endTime) {
		return (*current())[time];
	} else if (time < current()->startTime) {
		// find the right chunk, before the current one
		while (time < current()->startTime) {
			if (current.isHead()) {
				return 0.0;
			} else
				--current;
		}
	} else {
		// find the right chunk, after the current one
		while (time > current()->endTime) {
			if (current.isTail()) {
				time = current()->endTime;
				break;
				//return 0.0;
			} else
				++current;
		}
	}
	return (*current())[time];
}

//------------------------------------------------------------------------------------------------

void Track::TranslateTime(int offset, int start, int end)
{
	for(ListIterator<Chunk> c(data); c; ++c) {
		if (start<=c()->startTime && c()->endTime<=end) {
			c()->startTime += offset;
			c()->endTime += offset;
		}
	}
}

double Track::TranslateValue(double offset, int start, int end)
{
//fprintf(stderr, "\nTRANSLATE VALUE   %g  at  %d-%d",offset,start,end);

	for(ListIterator<Chunk> c(data); c; ++c) {
		if (start<=c()->startTime && c()->endTime<=end) {
			c()->startValue += offset;
			c()->leftValue += offset;
			c()->rightValue += offset;
			c()->endValue += offset;
		}
	}

	return -offset;
}

double Track::JointValue(int start, int end)
{
double offset = 0.0;

//fprintf(stderr, "\nJOINT INITIAL    ");for (ListIterator<Chunk> __c(data); __c; ++__c) {__c()->Dump(TRUE); fprintf(stderr, "\n        ");}


	// command to joint values
	ListIterator<Chunk> c(data);
	while (c) {
//fprintf(stderr, "\nsearch = %d-%d", c()->startTime, c()->endTime);
		if (c()->startTime >= start) {
			if (!c.isHead()) {
//fprintf(stderr, "\nJOINT [%d:%d] find = %g     ", start, end, -c.prev()->endValue+c()->startValue); c.prev()->Dump(TRUE);fprintf(stderr, " xx %g", c()->startValue);
				offset = c.prev()->endValue;
				// skip the synthetic chunk added to fill the gap : search real discontinuity
				while (c()->isSynthetic && !c.isTail() && c()->startValue == offset)
					++c;
				offset = c()->startValue - offset;
			} else {
//fprintf(stderr, "\nJOINT [%d:%d] find head = %g", start, end, c()->startValue);
				offset = c()->startValue;
			}
			break;
		}
		++c;
	}

//fprintf(stderr, "\n     offset = %g   starting chunk", offset); if (c) c()->Dump(TRUE);
	for(; c; ++c) {
		if (c()->endTime<=end) {
			c()->startValue -= offset;
			c()->leftValue -= offset;
			c()->rightValue -= offset;
			c()->endValue -= offset;
		}
	}

//fprintf(stderr, "\nJOINT FINAL    ");for (ListIterator<Chunk> __c(data); __c; ++__c) {__c()->Dump(TRUE); fprintf(stderr, "\n        ");}


	return offset;
}

//------------------------------------------------------------------------------------------------
// smooth a USED track
void Track::Smooth(int start, int end, int correct)
{
ListIterator<Chunk> c(data);

	// joint between 0 and start
	if (c()->startTime != 0) {
//fprintf(stderr, " start %d-%d", 0, c()->startTime);
		data->prepend(new Chunk(0, c()->startTime, c()->startValue,c()->startValue, 
								c()->startValue,c()->startValue, -1,-1, TRUE));
	}

	while (!c.isTail()) {
		// fill a possible gap
		if (c.next()->startTime != c()->endTime) {
			double startValue = c()->endValue;
			double endValue = c.next()->startValue;
			int startTime = c()->endTime;

//fprintf(stderr, " range=%d-%d ",start,end);

			// fill the gaps
			if (startTime < start && c.next()->startTime > start) {
//fprintf(stderr, " creating extra %d-%d", c()->endTime, start);
				startTime = start;
				data->insert(c, new Chunk(c()->endTime, start,
							 startValue, endValue, startValue+(endValue-startValue)/3,
							 startValue+(endValue-startValue)*2/3, -1,-1, TRUE));
				++c;
			}

//fprintf(stderr, " creating %d-%d", startTime, c.next()->startTime);
			data->insert(c, new Chunk(startTime, c.next()->startTime,
						 startValue, startValue, startValue, startValue, -1,-1, TRUE));

/*			data->insert(c, new Chunk(startTime, c.next()->startTime,
							 startValue, c.next()->startValue,
							 startValue+(c.next()->startValue-startValue)/3,
							 startValue+(c.next()->startValue-startValue)/2*3, -1,-1, TRUE));
*/
		}

		++c;
		// adjust values
		if (correct && !c.isTail() && c()->endValue != c.next()->startValue) {
//*** modified 98/8/2 after remove synthetic in figure

//			if (c()->isSynthetic) {
//fprintf(stderr, " correcting at %d>", c()->endTime);
				c.next()->startValue = c()->endValue;
				c.next()->leftValue = c.next()->startValue +
									(c.next()->endValue-c.next()->startValue)/3;
				c.next()->rightValue = c.next()->startValue + 
									 (c.next()->endValue-c.next()->startValue)*2/3;
/*			} else {
fprintf(stderr, " correcting at <%d", c()->endTime);
				c()->endValue = c.next()->startValue;
				c()->leftValue = c()->startValue + (c()->endValue-c()->startValue)/3;
				c()->rightValue = c()->startValue + (c()->endValue-c()->startValue)*2/3;
			}*/
		}
	}

	// handle the tail
	if (c()->endTime < end) {
//fprintf(stderr, " ending %d-%d", c()->endTime, end);
		data->insert(c, new Chunk(c()->endTime, end,
					 c()->endValue, c()->endValue, c()->endValue, c()->endValue, -1,-1, TRUE));
	}
}

//------------------------------------------------------------------------------------------------

void Track::RemoveSynthetic()
{
	for(ListIterator<Chunk> c(data); c; ++c) {
		c()->isSynthetic = FALSE;
	}
}

//------------------------------------------------------------------------------------------------

void Track::AdjustNbBeats(int nbBeats)
{
Chunk *c = data->last();

	if (c->startTime == c->endTime) {
		c->endTime = nbBeats;
	} else {
		data->append(new Chunk(c->endTime, nbBeats, 
							   c->endValue,c->endValue,c->endValue,c->endValue, -1,-1));
	}
}

//------------------------------------------------------------------------------------------------

int Track::LastDefinedTime(int time) const
{
ListIterator<Chunk> c(current);

	// if track no defined, returns 0
	if (!c) {
		return 0;
	}

	// if inside the current chunk
	if (c()->startTime <= time && time < c()->endTime) {
		;
	} else if (time < c()->startTime) {
		// find the right chunk, before the current one
		while (c()->startTime > time && !c.isHead()) {
			--c;
		}
	} else {
		// find the right chunk, after the current one
		while (c()->endTime <= time && !c.isTail()) {
			++c;
		}
	}

	return c()->startTime;
}

//------------------------------------------------------------------------------------------------

int Track::Update(int time, double newValue, int nbBeats, int mode, int lastTime)
{
enum {nothing, modifValue, newTrack, newKey};
int result;


//fprintf(stderr, "%08X  Track::Update @ %d = %g   --  %d - %d\n",this,time, newValue, nbBeats, mode);
	if (!current) {
//fprintf(stderr, "    no current\n");
		if ((time == 0 && newValue == 0.0) || newValue == 0.0) {
			return nothing;
		}

		if (mode == QUERY) {
//fprintf(stderr, "    query new track\n");
			return newTrack;
		}

		if (time) {
//fprintf(stderr, "new track : lastTime = %d\n", lastTime);
			if (lastTime != 0) {
				data->append(new Chunk(0, lastTime, 0,0,0,0, -1,-1));
			}
			data->append(new Chunk(lastTime, time, 0,newValue, newValue/3,newValue*2/3, -1,-1));
		} else {
			data->append(new Chunk(0, nbBeats, newValue,newValue,
								   newValue,newValue, -1,-1, TRUE));
		}

		current.reset(data);
		return newTrack;
	}

	FindPlace(time);

//fprintf(stderr, "    pos   = ");current()->Dump(TRUE);fprintf(stderr, "\n");

	if (current()->startTime > time) {
		// prepending a value
//fprintf(stderr, "!!!!!!!!!!!!! FATAL ERROR    <<prepending value>>\n");
		return newKey;
	}

	if (time == current()->startTime) {
		// modify value at beginning
//fprintf(stderr, "    <<modifying start>>\n");
		if (fabs(newValue - (*this)[time]) < 0.00001) {
			return nothing;
		}
		if (mode == QUERY) {
			return modifValue;
		}
		result = modifValue;
		current()->startValue = newValue;
		current()->SetLinear();
//fprintf(stderr, "    after modif end = \n");current()->Dump(TRUE);fprintf(stderr, "\n");
		if (current()->isSynthetic && current()->endTime == nbBeats) {
			current()->endValue = newValue;
			current()->SetLinear();
		} else {
			current()->isSynthetic = FALSE;
		}
		if (!current.isHead()) {
			current.prev()->endValue = newValue;
			current.prev()->SetLinear();
		}
	} else if (time == current()->endTime) {
		// modify value at end
//fprintf(stderr, "    <<modifying end>>\n");
		if (fabs(newValue - (*this)[time]) < 0.00001) {
			return nothing;
		}
		if (mode == QUERY) {
			return modifValue;
		}
		result = modifValue;
		current()->endValue = newValue;
		current()->SetLinear();
		current()->isSynthetic = FALSE;
//fprintf(stderr, "    after modif end = \n");current()->Dump(TRUE);fprintf(stderr, "\n");
		if (!current.isTail()) {
			current.next()->startValue = newValue;
			current.next()->SetLinear();
		}
	} else {
		// inserting or appending a new key
		if (fabs(newValue-(*this)[time]) < 0.00001 && mode != FORCE) {
//fprintf(stderr, "    <<same value --- doing nothing>>\n");
			return nothing;
		}
		if (mode == QUERY) {
			return newKey;
		}
		result = newKey;

		int endTime = current()->endTime;
		double endValue = current()->endValue;

		if (current()->startTime < time && time < endTime) {
			current()->endTime = time;
			current()->endValue = newValue;
			if (mode != APPLYMODIF) {
				current()->SetLinear();
			}
			int isSynthetic = current()->isSynthetic;
			current()->isSynthetic = FALSE;

			if (isSynthetic && endTime == nbBeats) {
//fprintf(stderr, "    <<inserting value with const synthetic end>>\n");
				data->insert(current, new Chunk(time, endTime, newValue, newValue,
							 newValue, newValue, -1,-1, TRUE));
			} else {
//fprintf(stderr, "    <<inserting value>>\n");
				data->insert(current, new Chunk(time, endTime, newValue, endValue,
							 newValue+(endValue-newValue)/3, newValue+(endValue-newValue)*2/3,
							 -1,-1));
			}
		} else {
			if (current()->endTime == 0) {
//fprintf(stderr, "    <<upgrading chunk end=0>>\n");
				current()->endTime = time;
				current()->isSynthetic = FALSE;
				current()->endValue = newValue;
				current()->SetLinear();
				return result;
			} else {
//fprintf(stderr, "    <<appending new value>>\n");
				data->append(new Chunk(endTime, time, endValue,newValue,
							 endValue+(newValue-endValue)/3, endValue+(newValue-endValue)*2/3,
							 -1,-1));
			}
		}
		++current;
	}

//fprintf(stderr, "    new t ="); if (IsUsed()) { for (ListIterator<Chunk> c(data); c; ++c) c()->Dump(TRUE); } fprintf(stderr, "\n");

	return result;
}


int Track::ApplyModif(class Modif *m, class List<Modif> *modifs)
{
ListIterator<Chunk> c;
int isEnd = FALSE, isHead = FALSE;

	// get the chunk to which apply the modification
restart:
	for(c.reset(data); c; ++c) {
		if (c()->startTime == m->time || (c.isTail() && c()->endTime == m->time)) {
			break;
		}
	}
	if (!c) {
//fprintf(stderr, "CREATE key via update\n");
		Update(m->time, m->data.value, 0, APPLYMODIF);
		m->type |= NEW_KEY;
		goto restart;
	}

	// check if we are dealing with the end
	isEnd = (c.isTail() && c()->endTime == m->time);
	isHead = (c.isHead());

	// if first modif, register old values
	if (!(m->type & OLD_VALUE) && (m->type & VAL_VALUE)) {
		m->oldData.value = isEnd ? c()->endValue : c()->startValue;
		m->type |= OLD_VALUE;
	}
	if (!(m->type & OLD_VALUE) && (m->type & VAL_LEFT)) {
		m->oldData.leftValue = isEnd ? c()->leftValue : isHead ? 0 : c.prev()->leftValue;
		m->oldData.leftTime = isEnd ? c()->leftTime : isHead ? -1 : c.prev()->leftTime;
		m->type |= OLD_LEFT;
	}
	if (!(m->type & OLD_VALUE) && (m->type & VAL_RIGHT)) {
		m->oldData.rightValue = isEnd ? c()->rightValue : isHead ? 0 : c.prev()->rightValue;
		m->oldData.rightTime = isEnd ? c()->rightTime : isHead ? -1 : c.prev()->rightTime;
		m->type |= OLD_RIGHT;
	}

	// apply it according to its type
//fprintf(stderr, "\nApplymodif : "); m->Dump();fprintf(stderr, "\n");
	if (m->type & VAL_VALUE) {
		if (isEnd) {
			c()->endValue = m->data.value;
		} else {
			c()->startValue = m->data.value;
			if (!isHead) {
				c.prev()->endValue = m->data.value;
			}
		}
	}

	class Chunk *cc;

	if (isEnd) {
		cc = c();
		if (m->type & VAL_RIGHT) {
//fprintf(stderr, "           -> VAL_RIGHT (end)\n");
			cc->rightValue = m->data.rightValue;
			cc->rightTime = m->data.rightTime;
		}
	} else {
		if (!isHead) {
			cc = c.prev();
			if (m->type & VAL_RIGHT) {
//fprintf(stderr, "           -> VAL_RIGHT (not head)\n");
				cc->rightValue = m->data.rightValue;
				cc->rightTime = m->data.rightTime;
			}
		}
		cc = c();
		if (m->type & VAL_LEFT) {
//fprintf(stderr, "           -> VAL_LEFT (current)\n");
			cc->leftValue = m->data.leftValue;
			cc->leftTime = m->data.leftTime;
		}
	}

	return TRUE;
}


//================================================================================================
//
//		An interpolator is an array of DOF tracks to memorize the movement
//
//================================================================================================

void Interpolator::SetBody(struct Body *b, int time, int joint)
{
	// update the value of joints (if joint specified, only this one, else all)
	if (joint >= 0) {
		if (track[joint].IsUsed()) {
			b->joints[joint] = track[joint][time];
		} else {
			b->joints[joint] = 0;
		}
	} else {
		for(int i=0; i<NB_DOF; i++) {
			if (track[i].IsUsed()) {
				b->joints[i] = track[i][time];
			} else {
				b->joints[i] = 0;
			}
		}
	}

	// force redisplay
	b->Redisplay();
}

void Interpolator::TranslateTime(int offset, int start, int end)
{
	for(int i=0; i<NB_DOF; i++) {
		if (track[i].IsUsed())
			track[i].TranslateTime(offset, start, end);
	}
}

void Interpolator::Smooth(int start, int end, int correct)
{
	for(int i=0; i<NB_DOF; i++) {
		if (track[i].IsUsed()) {
//fprintf(stderr, "\n %-20s = ",trackName[i]);
//if (i==cogX || i==cogZ || i==cogRY) for (ListIterator<Chunk> c(track[i].data); c; ++c) {fprintf(stderr,"\n            ");c()->Dump(TRUE);}

			track[i].Smooth(start, end, correct);
		}
	}
}

void Interpolator::RemoveSynthetic()
{
	for(int i=0; i<NB_DOF; i++) {
		if (track[i].IsUsed())
			track[i].RemoveSynthetic();
	}
}

void Interpolator::AdjustNbBeats(int nbBeats)
{
	for(int i=0; i<NB_DOF; i++) {
		if (track[i].IsUsed())
			track[i].AdjustNbBeats(nbBeats);
	}
}

//------------------------------------------------------------------------------------------------

#ifdef DEBUG
void Interpolator::Dump(int full) const
{
	for (int i=0; i<NB_DOF; i++) {
		if (track[i].IsUsed()) {
			fprintf(stderr, "    %s =", trackName[i]);
			for (ListIterator<Chunk> c(track[i].data); c; ++c) {
				fprintf(stderr,"\n            ");
				c()->Dump(full);
			}

			if (!full) {
				fprintf(stderr,"\n            ");
				track[i].data->last()->DumpEnd();
			}
			fprintf(stderr, "\n");
		}
	}
	fprintf(stderr, "\n");
}
#endif
